using Nemerle;
using Nemerle.Imperative;
using Nemerle.IO;
using System;
using System.IO;


namespace Rsdn.ParserTools.Bnf
{
  public class NodeBuilder : INodeBuilder
  {
    log : IErrorLog;
    
    public this(log : IErrorLog)
    {
      this.log = log;
    }
    
    public List(loc : TextLocator, head : Node, tail : Node) : Node
    {
      match (tail)
      {
        | Node.List(value) => Node.List(loc, head :: value)
        | Node.Null()      => Node.List(loc, head :: [])
        | _ => throw Exception($"Internal parser error at $loc")          
      }
    }
    
    ListToMap['a, 'b](node : Node, f : Node -> option['a * 'b], err : 'a -> string) : IConstListDictionary['a, 'b]
    {
      match (node)
      {
        | Node.List(value) =>
          def result = ListDictionary();
          foreach (child in value)
          {
            match (f(child))
            {
              | option.Some((a, b)) =>
                if (result.Keys.Contains(a))
                  log.Message(Error(child.loc, ErrorLevel.Error, err(a)))
                else
                  result.Add(a, b)
              | option.None() => ();
            }
          }
          result
        | Node.Null() => ListDictionary()
        | _ => throw Exception($"Internal parser error at $(node.loc)")
      }
    }
    
    public Root(loc : TextLocator, header : Node, ast : Node, subst : Node, lexers : Node, parser : Node) : Node
    {
      def headerMap = ListToMap(header,
        node =>
        {
          | Node.Arg(key, _) as arg => option.Some((key, arg))
          | _ => throw Exception($"Internal parser error at $loc")
        },
        key => $"Argument $key already defined");
        
      def astMap = ListToMap(ast,
        node =>
        {
          | Node.Constructor(id, _, _) as constructor => option.Some((id, constructor))
          | _ => throw Exception($"Internal parser error at $loc")
        },
        key => $"Constructor $key already defined");
      
      def substMap = ListToMap(subst,
        node =>
        {
          | Node.Rule(_, id, _, _) as rule => option.Some((id, rule))
          | _ => throw Exception($"Internal parser error at $loc")
        },
        key => $"Rule $key already defined");
      
      def lexerMap = ListToMap(lexers,
        node =>
        {
          | Node.Section(args, _) as lexer =>
            def id =
              if (args.Keys.Contains("name")) args["name"].value
            else
              "__main__";
            option.Some((id, lexer))
          | _ => throw Exception($"Internal parser error at $loc")
        },
        key => $"Lexer section $key already defined");
      
      def parserSection = match(parser)
      {
        | Node.Section(_, _) as section => section
        | Node.Null() => null
        | _ => throw Exception($"Internal parser error at $loc")
      }
      
      Node.Root(loc, headerMap, astMap, substMap, lexerMap, parserSection)
    }
    
    public Constructor(loc : TextLocator, isFunction : bool, idNode : Node, args : Node) : Node
    {
      match(idNode)
      {
        | Node.Id(id) =>
          def argsMap = ListToMap(args,
            node =>
            {
              | Node.Field(id, _) as field => option.Some((id, field))
              | _ => throw Exception("Internal parser error")
            },
            key => $"Constructor argument $key is already defined");
          Node.Constructor(loc, id, isFunction, argsMap)
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
    
    public ConstructorParam(loc : TextLocator, id : Node, typeId : Node) : Node
    {
      match ((id, typeId))
      {
        | (Node.Id(fieldId), Node.Id(fieldType)) => Node.Field(loc, fieldId, fieldType)
        | (Node.Id(fieldId), Node.Null()) => Node.Field(loc, fieldId, "node")
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
    
    public Section(loc : TextLocator, args : Node, body : Node) : Node
    {
      def argsMap = ListToMap(args,
        node =>
        {
          | Node.Arg(key, _) as arg => option.Some((key, arg))
          | _ => throw Exception("Internal parser error")
        },
        key => $"Argument $key already defined");
      
      def bodyMap = ListToMap(body,
        node =>
        {
          | Node.Rule(_, id, _,  _) as rule => option.Some((id, rule))
          | _ => throw Exception("Internal parser error")
        },
        key => $"Rule $key already defined");
      
      Node.Section(loc, argsMap, bodyMap)
    }
    
    public Arg(loc : TextLocator, key : Node, value : Node) : Node
    {
      match ((key, value))
      {
        | (Node.Id(argKey), Node.String(argValue)) => Node.Arg(loc, argKey, argValue)
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
    
    public Rule(loc : TextLocator, subst : bool, id : Node, tag : Node, expr : Node) : Node
    {
      match ((id, tag))
      {
        | (Node.Id(id), Node.Id(tag)) => Node.Rule(loc, subst, id, tag, expr)
        | (Node.Id(id), Node.Null()) => Node.Rule(loc, subst, id, null, expr)
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
    
    public Tagged(loc : TextLocator, expr : Node, tag : Node) : Node
    {
      match (tag)
      {
        | Node.Id(tag) => Node.Tagged(loc, expr, tag)
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
    
    public Range(loc : TextLocator, first : Node, second : Node) : Node
    {
      match ((first, second))
      {
        | (Node.String(first), Node.String(second)) =>
          if (first.Length != 1 || second.Length != 1)
          {
            log.Message(Error(loc, ErrorLevel.Error, "Operands of '..' operation must be single character strings"));
            Node.Range(loc, 0, 0)
          }
          else if (first[0] > second[0])
          {
            log.Message(Error(loc, ErrorLevel.Error, "The second operand of '..' operation must be greater than the first one"));
            Node.Range(loc, 0, 0)
          }
          else
            Node.Range(loc, first[0] :> int, second[0] :> int)
        | _ => throw Exception($"Internal parser error at $loc")
      }
    }
  }
}