
using Nemerle.Collections;
using Nemerle.Imperative;
using System;
using System.IO;
using System.IO.Compression;
using System.Text;
using Rsdn.ParserTools;

namespace Rsdn.ParserTools.Bnf
{

  public interface ITokenSource
  {
    GetNextToken() : Token;
    EvalToken(token : Token, log : IErrorLog) : Node;
  }
  
  public interface INodeBuilder
  {
    List(loc : TextLocator, head : Node, tail : Node) : Node;
  Root(loc : TextLocator, header : Node, ast : Node, subst : Node, lexers : Node, parser : Node) : Node;
  Constructor(loc : TextLocator, isFunction : bool, id : Node, args : Node) : Node;
  ConstructorParam(loc : TextLocator, id : Node, fieldType : Node) : Node;
  Section(loc : TextLocator, args : Node, body : Node) : Node;
  Arg(loc : TextLocator, key : Node, value : Node) : Node;
  Rule(loc : TextLocator, subst : bool, id : Node, tag : Node, expr : Node) : Node;
  Tagged(loc : TextLocator, expr : Node, tag : Node) : Node;
  Range(loc : TextLocator, first : Node, second : Node) : Node;
  
  }

  public class ParserHelper
  {
    private enum Action : byte
    {
      | Error
      | Shift
      | Reduce
      | Accept
    }

    static actionTable : array[array[int * Action * int]];
    static gotoTable : array[array[int * int]];
    mutable panic = false;

    static ProductionInfo(prod : int) : int * int
    {
      | 0 => (29, 2)
      | 1 => (67, 1)
      | 2 => (67, 0)
      | 3 => (68, 1)
      | 4 => (68, 0)
      | 5 => (69, 1)
      | 6 => (69, 0)
      | 7 => (70, 1)
      | 8 => (70, 0)
      | 9 => (71, 1)
      | 10 => (71, 0)
      | 11 => (30, 5)
      | 12 => (30, 5)
      | 13 => (31, 4)
      | 14 => (32, 2)
      | 15 => (33, 4)
      | 16 => (72, 1)
      | 17 => (72, 0)
      | 18 => (34, 2)
      | 19 => (73, 1)
      | 20 => (73, 0)
      | 21 => (35, 6)
      | 22 => (36, 3)
      | 23 => (36, 1)
      | 24 => (37, 3)
      | 25 => (37, 1)
      | 26 => (38, 4)
      | 27 => (74, 1)
      | 28 => (74, 0)
      | 29 => (39, 2)
      | 30 => (40, 1)
      | 31 => (40, 1)
      | 32 => (41, 3)
      | 33 => (41, 1)
      | 34 => (42, 3)
      | 35 => (43, 2)
      | 36 => (44, 7)
      | 37 => (44, 4)
      | 38 => (75, 1)
      | 39 => (75, 0)
      | 40 => (45, 2)
      | 41 => (46, 1)
      | 42 => (46, 1)
      | 43 => (47, 7)
      | 44 => (47, 4)
      | 45 => (76, 1)
      | 46 => (76, 0)
      | 47 => (48, 2)
      | 48 => (49, 1)
      | 49 => (49, 1)
      | 50 => (49, 1)
      | 51 => (50, 4)
      | 52 => (51, 5)
      | 53 => (52, 5)
      | 54 => (53, 1)
      | 55 => (53, 3)
      | 56 => (54, 1)
      | 57 => (54, 2)
      | 58 => (55, 1)
      | 59 => (55, 1)
      | 60 => (55, 1)
      | 61 => (56, 1)
      | 62 => (56, 2)
      | 63 => (57, 1)
      | 64 => (57, 2)
      | 65 => (58, 1)
      | 66 => (58, 2)
      | 67 => (59, 1)
      | 68 => (59, 1)
      | 69 => (60, 1)
      | 70 => (60, 3)
      | 71 => (61, 1)
      | 72 => (61, 3)
      | 73 => (62, 1)
      | 74 => (62, 2)
      | 75 => (63, 2)
      | 76 => (63, 1)
      | 77 => (64, 3)
      | 78 => (64, 1)
      | 79 => (64, 1)
      | 80 => (64, 1)
      | 81 => (65, 3)
      | 82 => (66, 3)
      | _ => throw Exception("Internal parser error")
    }

    static TokenToInt(token : TokenType) : int
    {
      | TokenType.Eof => 0
      | TokenType.Whitespace => 1
      | TokenType.Comment => 2
      | TokenType.Identifier => 3
      | TokenType.String => 4
      | TokenType.LeftBrace => 5
      | TokenType.RightBrace => 6
      | TokenType.LeftParen => 7
      | TokenType.RightParen => 8
      | TokenType.LeftBracket => 9
      | TokenType.RightBracket => 10
      | TokenType.Colon => 11
      | TokenType.Semicolon => 12
      | TokenType.Eq => 13
      | TokenType.Comma => 14
      | TokenType.Assign => 15
      | TokenType.VBar => 16
      | TokenType.Question => 17
      | TokenType.Plus => 18
      | TokenType.Star => 19
      | TokenType.Ellipsis => 20
      | TokenType.Amp => 21
      | TokenType.Backslash => 22
      | TokenType.Tilde => 23
      | TokenType.Fun => 24
      | TokenType.Lexer => 25
      | TokenType.Parser => 26
      | TokenType.Subst => 27
      | TokenType.Ast => 28
      | _ => throw Exception("Internal parser error")
    }

    static ReduceProduction(loc : TextLocator, builder : INodeBuilder, prod : int, stack : array[Node]) : Node
    {
      match (prod)
      {
        // copy: S_ -> Root Eof
        | 0 =>
          stack[0]
  
        // copy: Header_opt -> Header
        | 1 =>
          stack[0]
  
        // e-reduction: Header_opt -> 
        | 2 =>
          Node.Null(loc)
  
        // copy: AstSection_opt -> AstSection
        | 3 =>
          stack[0]
  
        // e-reduction: AstSection_opt -> 
        | 4 =>
          Node.Null(loc)
  
        // copy: SubstSection_opt -> SubstSection
        | 5 =>
          stack[0]
  
        // e-reduction: SubstSection_opt -> 
        | 6 =>
          Node.Null(loc)
  
        // copy: LexerSections_opt -> LexerSections
        | 7 =>
          stack[0]
  
        // e-reduction: LexerSections_opt -> 
        | 8 =>
          Node.Null(loc)
  
        // copy: ParserSection_opt -> ParserSection
        | 9 =>
          stack[0]
  
        // e-reduction: ParserSection_opt -> 
        | 10 =>
          Node.Null(loc)
  
        // reduction: Root -> Header_opt[header] AstSection_opt[ast] SubstSection_opt[subst] LexerSections_opt[lexers] ParserSection[parser]
        | 11 =>
          
        builder.Root(loc, stack[0], stack[1], stack[2], stack[3], stack[4])
      
        // reduction: Root -> Header_opt[header] AstSection_opt[ast] SubstSection_opt[subst] LexerSections[lexers] ParserSection_opt[parser]
        | 12 =>
          
        builder.Root(loc, stack[0], stack[1], stack[2], stack[3], stack[4])
      
        // reduction: Definition[Arg] -> Identifier[key] Eq String[value] Semicolon
        | 13 =>
          
        builder.Arg(loc, stack[0], stack[2])
      
        // reduction: Header[List] -> Definition[head] Header_opt[tail]
        | 14 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // copy: AstSection -> Ast LeftBrace AstDeclarations RightBrace
        | 15 =>
          stack[2]
  
        // copy: AstDeclarations_opt -> AstDeclarations
        | 16 =>
          stack[0]
  
        // e-reduction: AstDeclarations_opt -> 
        | 17 =>
          Node.Null(loc)
  
        // reduction: AstDeclarations[List] -> AstDeclaration[head] AstDeclarations_opt[tail]
        | 18 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // copy: Fun_opt -> Fun
        | 19 =>
          stack[0]
  
        // e-reduction: Fun_opt -> 
        | 20 =>
          Node.Null(loc)
  
        // reduction: AstDeclaration[Constructor] -> Fun_opt[isFunction] Identifier[id] LeftParen AstParams[args] RightParen Semicolon
        | 21 =>
          
        builder.Constructor(loc, ToBool(stack[0]), stack[1], stack[3])
      
        // reduction: AstParams[List] -> AstParam[head] Comma AstParams[tail]
        | 22 =>
          
        builder.List(loc, stack[0], stack[2])
      
        // reduction: AstParams[List] -> AstParam[head]
        | 23 =>
          
        builder.List(loc, stack[0], Node.Null(loc))
      
        // reduction: AstParam[ConstructorParam] -> Identifier[id] Colon Identifier[fieldType]
        | 24 =>
          
        builder.ConstructorParam(loc, stack[0], stack[2])
      
        // reduction: AstParam[ConstructorParam] -> Identifier[id]
        | 25 =>
          
        builder.ConstructorParam(loc, stack[0], Node.Null(loc))
      
        // copy: SubstSection -> Subst LeftBrace SubstDeclarations RightBrace
        | 26 =>
          stack[2]
  
        // copy: SubstDeclarations_opt -> SubstDeclarations
        | 27 =>
          stack[0]
  
        // e-reduction: SubstDeclarations_opt -> 
        | 28 =>
          Node.Null(loc)
  
        // reduction: SubstDeclarations[List] -> SubstDeclaration[head] SubstDeclarations_opt[tail]
        | 29 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // copy: SubstDeclaration -> Rule
        | 30 =>
          stack[0]
  
        // copy: SubstDeclaration -> Definition
        | 31 =>
          stack[0]
  
        // reduction: DefinitionArgs[List] -> DefinitionArg[head] Comma DefinitionArgs[tail]
        | 32 =>
          
        builder.List(loc, stack[0], stack[2])
      
        // reduction: DefinitionArgs[List] -> DefinitionArg[head]
        | 33 =>
          
        builder.List(loc, stack[0], Node.Null(loc))
      
        // reduction: DefinitionArg[Arg] -> Identifier[key] Eq String[value]
        | 34 =>
          
        builder.Arg(loc, stack[0], stack[2])
      
        // reduction: LexerSections[List] -> LexerSection[head] LexerSections_opt[tail]
        | 35 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // reduction: LexerSection[Section] -> Lexer LeftParen DefinitionArgs[args] RightParen LeftBrace LexerDeclarations[body] RightBrace
        | 36 =>
          
        builder.Section(loc, stack[2], stack[5])
      
        // reduction: LexerSection[Section] -> Lexer LeftBrace LexerDeclarations[body] RightBrace
        | 37 =>
          
        builder.Section(loc, Node.Null(loc), stack[2])
      
        // copy: LexerDeclarations_opt -> LexerDeclarations
        | 38 =>
          stack[0]
  
        // e-reduction: LexerDeclarations_opt -> 
        | 39 =>
          Node.Null(loc)
  
        // reduction: LexerDeclarations[List] -> LexerDeclaration[head] LexerDeclarations_opt[tail]
        | 40 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // copy: LexerDeclaration -> Rule
        | 41 =>
          stack[0]
  
        // copy: LexerDeclaration -> SubstRule
        | 42 =>
          stack[0]
  
        // reduction: ParserSection[Section] -> Parser LeftParen DefinitionArgs[args] RightParen LeftBrace ParserDeclarations[body] RightBrace
        | 43 =>
          
        builder.Section(loc, stack[2], stack[5])
      
        // reduction: ParserSection[Section] -> Parser LeftBrace ParserDeclarations[body] RightBrace
        | 44 =>
          
        builder.Section(loc, Node.Null(loc), stack[2])
      
        // copy: ParserDeclarations_opt -> ParserDeclarations
        | 45 =>
          stack[0]
  
        // e-reduction: ParserDeclarations_opt -> 
        | 46 =>
          Node.Null(loc)
  
        // reduction: ParserDeclarations[List] -> ParserDeclaration[head] ParserDeclarations_opt[tail]
        | 47 =>
          
        builder.List(loc, stack[0], stack[1])
      
        // copy: ParserDeclaration -> Rule
        | 48 =>
          stack[0]
  
        // copy: ParserDeclaration -> SubstRule
        | 49 =>
          stack[0]
  
        // copy: ParserDeclaration -> TaggedRule
        | 50 =>
          stack[0]
  
        // reduction: Rule[Rule] -> Identifier[id] Assign Union[expr] Semicolon
        | 51 =>
          
        builder.Rule(loc, false, stack[0], Node.Null(loc), stack[2])
      
        // reduction: SubstRule[Rule] -> Subst[subst] Identifier[id] Assign Union[expr] Semicolon
        | 52 =>
          
        builder.Rule(loc, ToBool(stack[0]), stack[1], Node.Null(loc), stack[3])
      
        // reduction: TaggedRule[Rule] -> Identifier[id] Tag[tag] Assign Union[expr] Semicolon
        | 53 =>
          
        builder.Rule(loc, false, stack[0], stack[1], stack[3])
      
        // copy: Union -> Concat
        | 54 =>
          stack[0]
  
        // reduction: Union -> Union[first] VBar Concat[second]
        | 55 =>
          Node.Union(loc, stack[0], stack[2])
      
        // copy: Concat -> ConcatExpr
        | 56 =>
          stack[0]
  
        // reduction: Concat -> Concat[first] ConcatExpr[second]
        | 57 =>
          Node.Concat(loc, stack[0], stack[1])
      
        // copy: ConcatExpr -> Optional
        | 58 =>
          stack[0]
  
        // copy: ConcatExpr -> Closure
        | 59 =>
          stack[0]
  
        // copy: ConcatExpr -> PositiveClosure
        | 60 =>
          stack[0]
  
        // copy: Optional -> CharExpr
        | 61 =>
          stack[0]
  
        // reduction: Optional -> ConcatExpr[expr] Question
        | 62 =>
          Node.Optional(loc, stack[0])
      
        // copy: Closure -> CharExpr
        | 63 =>
          stack[0]
  
        // reduction: Closure -> ConcatExpr[expr] Star
        | 64 =>
          Node.Closure(loc, stack[0])
      
        // copy: PositiveClosure -> CharExpr
        | 65 =>
          stack[0]
  
        // reduction: PositiveClosure -> ConcatExpr[expr] Plus
        | 66 =>
          Node.PositiveClosure(loc, stack[0])
      
        // copy: CharExpr -> Intersection
        | 67 =>
          stack[0]
  
        // copy: CharExpr -> Difference
        | 68 =>
          stack[0]
  
        // copy: Intersection -> Complement
        | 69 =>
          stack[0]
  
        // reduction: Intersection -> CharExpr[first] Amp Complement[second]
        | 70 =>
          Node.Intersection(loc, stack[0], stack[2])
      
        // copy: Difference -> Complement
        | 71 =>
          stack[0]
  
        // reduction: Difference -> CharExpr[first] Backslash Complement[second]
        | 72 =>
          Node.Difference(loc, stack[0], stack[2])
      
        // copy: Complement -> Tagged
        | 73 =>
          stack[0]
  
        // reduction: Complement -> Tilde Complement[expr]
        | 74 =>
          Node.Complement(loc, stack[1])
      
        // reduction: Tagged -> Primitive[expr] Tag[tag]
        | 75 =>
          
        builder.Tagged(loc, stack[0], stack[1])
      
        // copy: Tagged -> Primitive
        | 76 =>
          stack[0]
  
        // copy: Primitive -> LeftParen Union RightParen
        | 77 =>
          stack[1]
  
        // copy: Primitive -> String
        | 78 =>
          stack[0]
  
        // copy: Primitive -> Range
        | 79 =>
          stack[0]
  
        // copy: Primitive -> Identifier
        | 80 =>
          stack[0]
  
        // copy: Tag -> LeftBracket Identifier[_] RightBracket
        | 81 =>
          stack[1]
  
        // reduction: Range -> String[first] Ellipsis String[second]
        | 82 =>
          
        builder.Range(loc, stack[0], stack[2])
      
        | _ => throw Exception("Internal parser error")
      }
    }

    static ToBool(node : Node) : bool
    {
      | Node.Null() => false
      | _ => true
    }

    static this()
    {
      def stream = typeof(ParserHelper).Assembly.GetManifestResourceStream(typeof(ParserHelper), "Parser.bin");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def stateCount = Int32.Parse(reader.ReadLine());
      actionTable = array(stateCount);
      gotoTable = array(stateCount);
      for (mutable i = 0; i < stateCount; ++i)
      {
        def cellCount = Int32.Parse(reader.ReadLine());
        def row = array(cellCount);
        for (mutable j = 0; j < cellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def action = Int32.Parse(reader.ReadLine()) :> Action;
          def param = Int32.Parse(reader.ReadLine());
          row[j] = (symbol, action, param);
        }
        actionTable[i] = row;
        def gotoCellCount = Int32.Parse(reader.ReadLine());
        def gotoRow = array(gotoCellCount);
        for (mutable j = 0; j < gotoCellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def param = Int32.Parse(reader.ReadLine());
          gotoRow[j] = (symbol, param);
        }
        gotoTable[i] = gotoRow;
      }
    }

    public Panic : bool
    {
      get
      {
        return panic;
      }
    }

    public ResetPanic() : void
    {
      panic = false;
    }        

    public Parse(source : ITokenSource, builder : INodeBuilder, log : IErrorLog) : Node
    {
      mutable token;
      def stack = Vector.[int * int * Node * Token * Token]();
      def next_token()
      {
        token = source.GetNextToken();
      }
      def state()
      {
        if (stack.Count > 0)
          stack[stack.Count - 1][0]
        else
          0
      }
      def search_array['a, 'b](e : 'a, arr : array['b], comparer)
      {
        def search(l, u)
        {
          def i = (l + u) / 2;
          if (l > u)
            -1
          else
          {
            match (comparer(e, arr[i]))
            {
              | -1  => search(l, i - 1)
              | 1   => search(i + 1, u)
              | _   => i
            }
          }
        }
        search(0, arr.Length - 1)
      }
      def get_action(state, symbol)
      {
        def row = actionTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        if (col >= 0) (row[col][1], row[col][2]) else (Action.Error, -1)
      }
      def get_goto(state, symbol)
      {
        def row = gotoTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        row[col][1]
      }
      def shift(state)
      {
        def symbol = TokenToInt(token.TokenType);
        def data = source.EvalToken(token, log);
        stack.Add((state, symbol, data, token, token));
        next_token()
      }
      def reduce(prodIndex)
      {
        def (nonterm, len) = ProductionInfo(prodIndex);
        def get_data(i, acc)
        {
          if (i < 0)
            acc
          else
          {
            def (_, _, data, _, _) = stack[stack.Count - len + i];
            get_data(i - 1, data :: acc)
          }
        }
        def data = get_data(len - 1, []);
        def (firstToken, lastToken) =
          if (len > 0)
            (stack[stack.Count - len][3], stack[stack.Count - 1][4])
          else
            (token, token);
        def loc = TextLocator(firstToken.Locator.Url, TextRange(firstToken.Locator.First, lastToken.Locator.Last));
        def newData = ReduceProduction(loc, builder, prodIndex, data.ToArray());
        stack.RemoveRange(stack.Count - len, len);
        def newState = get_goto(state(), nonterm);
        stack.Add((newState, nonterm, newData, firstToken, lastToken));
      }
      def error()
      {
        def search_state(i)
        {
          if (i < -1)
            i
          else
          {
            def testState = if (i > 0) stack[i][0] else 0;
            def (action, _) = get_action(testState, TokenToInt(token.TokenType));
            match (action)
            {
              | Action.Error => search_state(i - 1)
              | _            => i
            }
          }
        }
        panic = true;
        mutable firstToken = token;
        while (true)
        {
          def stackIndex = search_state(stack.Count - 1);
          if (stackIndex >= -1)
          {
            firstToken = if (stackIndex + 1 < stack.Count)
              stack[stackIndex + 1][3]
            else
              token;
            stack.RemoveRange(stackIndex + 1, stack.Count - stackIndex - 1);
            break;
          }
          else
            next_token();
          when (token.TokenType == TokenType.Eof) break;
        }
        def loc = TextLocator(firstToken.Locator.Url, firstToken.Locator.First, token.Locator.Last);
        log.Message(Error(loc, ErrorLevel.Error, "Invalid syntax"));
        token.TokenType != TokenType.Eof
      }
      def loop()
      {
        def (action, param) = get_action(state(), TokenToInt(token.TokenType));
        match (action)
        {
          | Action.Shift  => shift(param); loop()
          | Action.Reduce => reduce(param); loop()
          | Action.Accept => stack[0][2]
          | Action.Error  => if (error()) loop() else stack[0][2]
        }
      }
      next_token();
      loop()
    }
  }

  public class LexerHelper
  {
    static stateNames : array[array[TokenType]] = array[
      array[
        TokenType.Error,
        TokenType.Comment,
        TokenType.Whitespace,
        TokenType.Error,
        TokenType.Amp,
        TokenType.LeftParen,
        TokenType.RightParen,
        TokenType.Star,
        TokenType.Plus,
        TokenType.Comma,
        TokenType.Error,
        TokenType.Error,
        TokenType.Colon,
        TokenType.Semicolon,
        TokenType.Eq,
        TokenType.Question,
        TokenType.Error,
        TokenType.Identifier,
        TokenType.LeftBracket,
        TokenType.Backslash,
        TokenType.RightBracket,
        TokenType.LeftBrace,
        TokenType.VBar,
        TokenType.RightBrace,
        TokenType.Tilde,
        TokenType.Whitespace,
        TokenType.Error,
        TokenType.String,
        TokenType.Error,
        TokenType.Ellipsis,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Identifier,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Comment,
        TokenType.Error,
        TokenType.Assign,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Comment,
        TokenType.Error,
        TokenType.Error,
        TokenType.Error,
        TokenType.Comment,
        TokenType.Error,
        TokenType.Error,

      ]
    ];
    static automata : array[array[array[int * int * int]]];
    static initialState : int * int * TokenType = (0, 0, TokenType.Eof);
    reader : TextReader;
    strBuilder : StringBuilder = StringBuilder();
    url : string;
    mutable lastToken : TokenType;
    mutable mode : int;
    mutable ch : int;
    mutable row : int;
    mutable column : int;
    mutable currentState : int;
    mutable cr : bool;
        
    static this()
    {
      def stream = typeof(LexerHelper).Assembly.GetManifestResourceStream(typeof(LexerHelper), "Lexer.bin");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def automataCount = Int32.Parse(reader.ReadLine());
      automata = array(automataCount);
      for (mutable index = 0; index < automataCount; ++index)
      {
        def stateCount = Int32.Parse(reader.ReadLine());
        def states = array(stateCount);
        for (mutable n = 0; n < stateCount; ++n)
        {
          def transitionsCount = Int32.Parse(reader.ReadLine());
          def transitions = array(transitionsCount);
          for (mutable i = 0; i < transitionsCount; ++i)
          {
            def from = Int32.Parse(reader.ReadLine());
            def to = Int32.Parse(reader.ReadLine());
            def state = Int32.Parse(reader.ReadLine());
            transitions[i] = (from, to, state);
          }
          states[n] = transitions;
        }
        automata[index] = states;
      }
      reader.Close();
    }
        
    public this(reader : TextReader, url : string)
    {
       this.reader = reader;
       this.url = url;
       ch = 0;
       column = -1;
       get_next_char();
    }
        
    public static InitialState : object
    {
      get
      {
        return initialState;
      }
    }
        
    get_next_char() : void
    {
      if (ch == 13)
      {
        cr = true;
        ++row;
        column = 0;
      }
      else if (ch == 10)
      {
        column = 0;
        when (!cr) ++row;
        cr = false;
      }
      else
      {
        ++column;
        cr = false;
      }
      ch = reader.Read();
    }
        
    public Mode : int
    {
      get
      {
        return this.mode;
      }
      set
      {
        this.mode = value;
      }
    }
        
    public LastTokenType : TokenType
    {
      get
      {
        return this.lastToken;
      }
    }
        
    public IsEof : bool
    {
      get
      {
        return ch == -1;
      }
    }
        
    public Save() : object
    {
      (currentState, mode, lastToken);
    }
        
    public Reset(state : object, row : int, column : int) : void
    {
      def extState = state :> int * int * TokenType;
      this.currentState = extState[0];
      this.lastToken = extState[2];
      this.mode = extState[1];
      this.row = row;
      this.column = column;
    }
    
    TransformToken(token : TokenType, text : string) : TokenType
    {
      match (mode)
      {
        
        | 0 =>
          match (token)
          {
            
            | TokenType.Identifier =>
              match (text)
              {
                
                | "fun" => TokenType.Fun
                | "lexer" => TokenType.Lexer
                | "parser" => TokenType.Parser
                | "subst" => TokenType.Subst
                | "ast" => TokenType.Ast
                | _ => token
              }
  
            | _ => token
          }
  
        | _ => token
      }
    }
    
    public GetToken() : Token
    {
      def get_next_state(mode, state, c)
      {
        def s = automata[mode][state];
        def search(l, u)
        {
          if (l > u)
            -1
          else
          {
            def i = (l + u) / 2;
            if (c < s[i][0])
              search(l, i - 1)
            else if (c > s[i][1])
              search(i + 1, u)
            else
              s[i][2]
          }
        }
        search(0, s.Length - 1)
      }
      if (ch == -1)
      {
         Token(
         TokenType.Eof,
          TextLocator(url, TextRange(row, column, row, column)),
          string.Empty)
      }
      else
      {
        def firstRow = row;
        def firstColumn = column;
        mutable lastRow = row;
        mutable lastColumn = column;
        _ = strBuilder.Remove(0, strBuilder.Length);
        mutable lastState = currentState;
        while (true)
        {
          lastRow = row;
          lastColumn = column;
          def nextState = get_next_state(mode, currentState, ch);
          when (nextState == -1) break;
          currentState = nextState;
          when (ch == -1) break;
          lastState = nextState;
          _ = strBuilder.Append(ch :> char);
          get_next_char();
        }
        when (stateNames[mode][currentState] == TokenType.Error)
        {
          get_next_char();
        }
        lastToken = stateNames[mode][currentState];
        if (ch == -1)
        {
          currentState = lastState;
        }
        else
        {
          currentState = 0;
        }
        def text = strBuilder.ToString();
        Token(
          TransformToken(lastToken, text),
          TextLocator(url, TextRange(firstRow, firstColumn, lastRow, lastColumn)),
          text);
      }
    }
  }
  
  public class Lexer : ITokenSource
  {
    helper : LexerHelper;
    
    public this(reader : TextReader, url : string)
    {
      this.helper = LexerHelper(reader, url);
    }
    
    public GetNextToken() : Token
    {
      def loop(token)
      {
        match (token.TokenType)
        {
          | TokenType.Eof => token
          | _ when TokenManager.Filter(token) => token
          | _ => loop(helper.GetToken())
        }
      }
      loop(helper.GetToken())
    }
    
    public EvalToken(token : Token, log : IErrorLog) : Node
    {
      TokenManager.Eval(token, log)
    }
  }
  
  [Record]
  public class Token
  {
    public TokenType : TokenType;
    public Locator : TextLocator;
    public Text : string;

    public this(proto : Token)
    {
      this.TokenType = proto.TokenType;
      this.Text = proto.Text;
      this.Locator = proto.Locator;
    }
        
    public this(proto : Token, tokenType : TokenType)
    {
      this.TokenType = tokenType;
      this.Text = proto.Text;
      this.Locator = proto.Locator;
    }
  }
  
  public enum TokenType
  {
    | Whitespace
    | Comment
    | Identifier
    | String
    | LeftBrace
    | RightBrace
    | LeftParen
    | RightParen
    | LeftBracket
    | RightBracket
    | Colon
    | Semicolon
    | Eq
    | Comma
    | Assign
    | VBar
    | Question
    | Plus
    | Star
    | Ellipsis
    | Amp
    | Backslash
    | Tilde
    | Fun
    | Lexer
    | Parser
    | Subst
    | Ast

    | Eof
    | Error
  }

}