using Nemerle;
using Nemerle.Collections;
using Nemerle.Utility;
using System.Globalization;
using Rsdn.ParserTools.Grammar;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools.Bnf
{
  public module BnfUtils
  {
    public PredefinedRegexps : IConstListDictionary[string, RegexpNode];
    
    this()
    {
      def unicode_node(category)
      {
        RegexpNode.Range(CharRange.GetUnicodeRange(category))
      }
      def pre = ListDictionary();
      pre.Add("U_Pe", unicode_node(UnicodeCategory.ClosePunctuation));
      pre.Add("U_Pc", unicode_node(UnicodeCategory.ConnectorPunctuation));
      pre.Add("U_Cc", unicode_node(UnicodeCategory.Control));
      pre.Add("U_Sc", unicode_node(UnicodeCategory.CurrencySymbol));
      pre.Add("U_Pd", unicode_node(UnicodeCategory.DashPunctuation));
      pre.Add("U_Nd", unicode_node(UnicodeCategory.DecimalDigitNumber));
      pre.Add("U_Me", unicode_node(UnicodeCategory.EnclosingMark));
      pre.Add("U_Pf", unicode_node(UnicodeCategory.FinalQuotePunctuation));
      pre.Add("U_Cf", unicode_node(UnicodeCategory.Format));
      pre.Add("U_Pi", unicode_node(UnicodeCategory.InitialQuotePunctuation));
      pre.Add("U_Nl", unicode_node(UnicodeCategory.LetterNumber));
      pre.Add("U_Zl", unicode_node(UnicodeCategory.LineSeparator));
      pre.Add("U_Ll", unicode_node(UnicodeCategory.LowercaseLetter));
      pre.Add("U_Sm", unicode_node(UnicodeCategory.MathSymbol));
      pre.Add("U_Lm", unicode_node(UnicodeCategory.ModifierLetter));
      pre.Add("U_Sk", unicode_node(UnicodeCategory.ModifierSymbol));
      pre.Add("U_Mn", unicode_node(UnicodeCategory.NonSpacingMark));
      pre.Add("U_Ps", unicode_node(UnicodeCategory.OpenPunctuation));
      pre.Add("U_Lo", unicode_node(UnicodeCategory.OtherLetter));
      pre.Add("U_Cn", unicode_node(UnicodeCategory.OtherNotAssigned));
      pre.Add("U_No", unicode_node(UnicodeCategory.OtherNumber));
      pre.Add("U_Po", unicode_node(UnicodeCategory.OtherPunctuation));
      pre.Add("U_So", unicode_node(UnicodeCategory.OtherSymbol));
      pre.Add("U_Zp", unicode_node(UnicodeCategory.ParagraphSeparator));
      pre.Add("U_Co", unicode_node(UnicodeCategory.PrivateUse));
      pre.Add("U_Zs", unicode_node(UnicodeCategory.SpaceSeparator));
      pre.Add("U_Mc", unicode_node(UnicodeCategory.SpacingCombiningMark));
      pre.Add("U_Cs", unicode_node(UnicodeCategory.Surrogate));
      pre.Add("U_Lt", unicode_node(UnicodeCategory.TitlecaseLetter));
      pre.Add("U_Lu", unicode_node(UnicodeCategory.UppercaseLetter));
      pre.Add("eof", RegexpNode.Range(CharRange.Eof));
      pre.Add("any", RegexpNode.Range(CharRange.Full));
      pre.Add("empty", RegexpNode.Empty());
      PredefinedRegexps = pre;
    }
    
    /*
        Applies all substitution in AST
     */
    public ApplySubst(node : Node.Root, log : IErrorLog) : Node.Root
    {
      // apply substitutions in lexers
      //
      def lexers = ListDictionary();
      def subst = node.subst;
      def parserPredefined = ListSet();
      foreach (id in node.lexers.Keys)
      {
        def lexer = node.lexers[id];
        def rules = ListDictionary();
        def lexerSubst = ListDictionary();
        foreach (id in subst.Keys)
          lexerSubst.Add(id, subst[id]);
        foreach (ruleId in lexer.body.Keys)
        {
          def rule = lexer.body[ruleId];
          if (lexerSubst.Keys.Contains(ruleId))
            log.Message(Error(rule.loc, ErrorLevel.Error, $"Subst section already contains rule named $ruleId"))
          else if (rule.subst)
            lexerSubst.Add(ruleId, rule)
          else
            rules.Add(ruleId, rule)
        }
        parserPredefined.AddMany(rules.Keys);
        def appliedLexer = Node.Section(
          lexer.loc, lexer.args,
          ApplySubst(rules, lexerSubst, PredefinedRegexps.Keys, log));
        lexers.Add(id, appliedLexer)
      }
      
      // apply substitions in parser
      //
      def parserSubst = ListDictionary();
      def parserRules = ListDictionary();
      when (node.parser != null)
      {
        foreach (id in node.parser.body.Keys)
        {
          def rule = node.parser.body[id];
          if (rule.subst)
            parserSubst.Add(id, rule)
          else
            parserRules.Add(id, rule)
        }
      }
      def appliedParser = if (node.parser == null)
        null
      else
        Node.Section(
          node.parser.loc, node.parser.args,
          ApplySubst(parserRules, parserSubst, parserPredefined, log));
      
      Node.Root(node.loc, node.header, node.ast, ListDictionary(), lexers, appliedParser)
    }
    
    /*
        Applies substitutions to single rule list and checks for invalid identifiers
     */
    public ApplySubst(
      rules : IConstListDictionary[string, Node.Rule],
      subst : IConstListDictionary[string, Node.Rule],
      predefined : IConstListSet[string],
      log : IErrorLog) : IConstListDictionary[string, Node.Rule]
    {
      def error(node, msg)
      {
        log.Message(Error(node.loc, ErrorLevel.Error, msg))
      }
      def processed = ListDictionary();
      def result = ListDictionary();
      
      def make_node(node)
      {
        def visit_id(node, id, walk, visited)
        {
          // preserve visited set for siblings
          def localvisited = HashSet(visited);
          if (predefined.Contains(id) || rules.Keys.Contains(id))
            node
          else if (processed.Keys.Contains(id))
            processed[id]
          else if (subst.Keys.Contains(id))
          {
            def result = if (localvisited.Contains(id))
            {
              error(node, $"Recursive reference found: $id");
              log.SetPanic();
              Node.Null(node.loc)
            }
            else
            {
              localvisited.Add(id);
              walk(subst[id].expr, localvisited);
            }
            unless (processed.Keys.Contains(id))
              processed.Add(id, result);
            result
          }
          else
          {
            error(node, $"Undeclared identifier $id");
            log.SetPanic();
            Node.Null(node.loc)
          }
        }
        def loop(node, visited)
        {
          match (node)
          {
            | Node.Union(first, last) =>
              Node.Union(node.loc, loop(first, visited), loop(last, visited))
            | Node.Concat(first, last) =>
              Node.Concat(node.loc, loop(first, visited), loop(last, visited))
            | Node.Optional(expr) =>
              Node.Optional(node.loc, loop(expr, visited))
            | Node.Closure(expr) =>
              Node.Closure(node.loc, loop(expr, visited))
            | Node.PositiveClosure(expr) =>
              Node.PositiveClosure(node.loc, loop(expr, visited))
            | Node.Intersection(first, last) => 
              Node.Intersection(node.loc, loop(first, visited), loop(last, visited))
            | Node.Difference(first, last) =>
              Node.Difference(node.loc, loop(first, visited), loop(last, visited))
            | Node.Complement(expr) =>
              Node.Complement(node.loc, loop(expr, visited))
            | Node.Id(id) => visit_id(node, id, loop, visited)
            | Node.Tagged(expr, id)    => Node.Tagged(node.loc, loop(expr, visited), id)
            | Node.Range(first, last)  => Node.Range(node.loc, first, last)
            | Node.String(value)       => Node.String(node.loc, value)
            | _                        => error(node, "Internal error"); Node.Null(node.loc)
          }
        }
        loop(node, HashSet())
      }
      
      foreach (key in rules.Keys)
      {
        def raw = rules[key];
        result.Add(key, Node.Rule(raw.loc, false, key, raw.tag, make_node(raw.expr)));
      }
      result
    }
    
    public GetTerminals(root : Node.Root) : SCG.IEnumerable[string]
    {
      foreach (lexerId in root.lexers.Keys)
      {
        def lexer = root.lexers[lexerId];
        foreach (id when !lexer.body[id].subst in lexer.body.Keys)
          yield id;
      }
    }
    
    public GetNonterminals(root : Node.Root) : SCG.IEnumerable[string]
    {
      root.parser.body.Keys;
    }
    
    public BuildGrammar(root : Node.Root, log : IErrorLog) : Grammar
    {
      def productions = ListMultiDictionary();
      def terminals = ListSet();
      def nonterminals = ListSet();
      def symbols = SCG.Dictionary.[string, Symbol]();
      def strmap = SCG.Dictionary.[string, Symbol]();
      
      def error(node, message)
      {
        log.Message(Error(node.loc, ErrorLevel.Error, message))
      }
      
      def build_prods(node) : SCG.IEnumerable[list[Symbol * string * TextLocator]]
      {
        def handle_string(value, tag, node)
        {
          if (strmap.ContainsKey(value))
            [[(strmap[value], tag, node.loc)]]
          else
          {
            error(node, $"The '$value' does not match any token class of single string");
            []
          }
        }
        
        def handle_optional(expr, tag)
        {
          def make_opt(name, tag, loc)
          {
            def optname = name + "_opt";
            def nonterm = if (symbols.ContainsKey(optname))
              symbols[optname] :> Nonterminal
            else
            {
              def r = ConcreteNonterminal(optname);
              symbols.Add(optname, r);
              nonterminals.Add(r);
              r
            }
            def prod = Production(nonterm, [(symbols[name], null, loc)], null, loc);
            productions.Add(nonterm, prod);
            def eprod = Production(nonterm, [], null, loc);
            productions.Add(nonterm, eprod);
            [(nonterm, tag, loc)]
          }
          match (expr)
          {
            | Node.Id(name) => [make_opt(name, tag, expr.loc)];
            | Node.String(value) =>
              if (strmap.ContainsKey(value))
                [make_opt(strmap[value].Name, tag, expr.loc)]
              else
              {
                error(expr, $"The '$value' does not match any token class of single string");
                []
              }
            | Node.Tagged(Node.Id(id), tag) => [make_opt(id, tag, expr.loc)]
            | other =>
              def res = Vector(build_prods(other));
              res.Add([]);
              res
          }
        }
        
        def build_tagged_expr(node, tag) : SCG.IEnumerable[list[Symbol * string * TextLocator]]
        {
          match (node)
          {
            | Node.Union(first, second) =>
              def res = Vector();
              res.AddRange(build_tagged_expr(first, tag));
              res.AddRange(build_tagged_expr(second, tag));
              res
            | Node.Id(id) => [[(symbols[id], tag, node.loc)]]
            | Node.String(value) => handle_string(value, tag, node)
            | Node.Optional(expr) => handle_optional(expr, tag)
            | _ => error(node, "Expression cannot have tag"); []
          }
        }
        
        match(node)
        {
          | Node.Union(first, second) =>
            def res = Vector();
            res.AddRange(build_prods(first));
            res.AddRange(build_prods(second));
            res
          | Node.Concat(first, second) =>
            def res = Vector();
            foreach (a in build_prods(first))
              foreach (b in build_prods(second))
                res.Add(a + b);
            res
          | Node.Optional(expr) => handle_optional(expr, null)
          | Node.Id(name) => [[(symbols[name], null, node.loc)]]
          | Node.Tagged(expr, tag) => build_tagged_expr(expr, tag)
          | Node.Range(_, _) => error(node, "Operation '..' is not valid for this section"); []
          | Node.String(value) => handle_string(value, null, node)
          | Node.Intersection(_, _) => error(node, "Operation '&' is not valid for this section"); []
          | Node.Difference(_, _) => error(node, "Operation '\\' is not valid for this section"); []
          | Node.Complement(_) => error(node, "Operation '~' is not valid for this section"); []
          | Node.Closure(_) => error(node, "Operation '*' not yet supported"); []
          | Node.PositiveClosure(_) => error(node, "Operation '+' not yet supported"); []
          | _ => error(node, "Invalid ast or internal bug"); []
        }
      }
      
      def fill_symbols()
      {
        symbols.Add("___eof___", EofTerminal.Instance);
        terminals.Add(EofTerminal.Instance);
        symbols.Add("___s___", StartNonterminal.Instance);
        nonterminals.Add(StartNonterminal.Instance);
        foreach (name when !symbols.ContainsKey(name) in GetTerminals(root))
        {
          def symbol = ConcreteTerminal(name);
          terminals.Add(symbol);
          symbols.Add(name, symbol)
        }
        foreach (name in GetNonterminals(root))
        {
          if (symbols.ContainsKey(name))
            log.Message(Error(root.parser.body[name].loc, ErrorLevel.Error, $"Rule $name already defined in a lexer section"));
          else
          {
            def symbol = ConcreteNonterminal(name);
            nonterminals.Add(symbol);
            symbols.Add(name, symbol)
          }
        }
      }
      
      def fill_strmap()
      {
        def tmpstrings = ListSet();
        foreach (lexerId in root.lexers.Keys)
        {
          def lexer = root.lexers[lexerId];
          def add(value, id)
          {
            if (tmpstrings.Contains(value))
            {
              when (strmap.ContainsKey(value)) _ = strmap.Remove(value);
            }
            else
            {
              strmap.Add(value, symbols[id]);
              tmpstrings.Add(value);
            }
          }
          foreach (id in lexer.body.Keys)
          {
            match (lexer.body[id].expr)
            {
              | Node.String(value) => add(value, id)                
              | _ => ();
            }
          }
        }
      }
      
      def fill_prods()
      {
        def rootSymbol = if (root.parser.args.Keys.Contains("root"))
        {
          def name = root.parser.args["root"].value;
          if (symbols.ContainsKey(name)) symbols[name] else null
        }
        else
          null;
          
        if (rootSymbol == null)
        {
          error(root.parser, "Root symbol was not defined");
          log.SetPanic()
        }
        else
        {
          def rootLoc = root.parser.body[rootSymbol.Name].loc;
          def firstProd = Production(
            StartNonterminal.Instance,
            [(rootSymbol, null, rootLoc), (EofTerminal.Instance, null, rootLoc)],
            null, rootLoc);
          productions.Add(StartNonterminal.Instance, firstProd);
          foreach (id in root.parser.body.Keys)
          {
            def rule = root.parser.body[id];
            def nonterm = symbols[id] :> Nonterminal;
            def prods = productions[nonterm];
            foreach (sentence in build_prods(rule.expr))
            {
              def prod = Production(nonterm, sentence, rule.tag, rule.loc);
              if (prods.Contains(prod))
              {
                when (!Production.SameTags(prods.GetOriginal(prod), prod))
                {
                  log.Message(Error(
                    rule.expr.loc,
                    ErrorLevel.Error,
                    "There are same production variants with different tags"))
                }
              }
              else
                productions.Add(nonterm, prod)
            }
          }
        }
      }
      
      fill_symbols();
      fill_strmap();
      fill_prods();
      Grammar(terminals, nonterminals, productions)
    }
    
    public BuildRegexps(lexer : Node.Section, log : IErrorLog) : IConstListDictionary[string, RegexpNode]
    {
      def rules = ListDictionary();
      foreach (id in lexer.body.Keys)
        rules.Add(id, lexer.body[id].expr);
      BuildRegexps(rules, log)
    }
    
    public BuildRegexps(
      rules : IConstListDictionary[string, Node],
      log : IErrorLog) : IConstListDictionary[string, RegexpNode]
    {
      def int_err = "The '&' operation is only applicable to character arguments";
      def diff_err = "The '\\' operation is only applicable to character arguments";
      def compl_err = "The '~' operation is only applicable to character argument";
      def error(node, message)
      {
        log.Message(Error(node.loc, ErrorLevel.Error, message));
      }
      
      def build_concat(first, second)
      {
        | (RegexpNode.Empty(), b) => b
        | (a, RegexpNode.Empty()) => a
        | (a, b) => RegexpNode.Concat(a, b)
      }
      def build_union(first, second)
      {
        | (RegexpNode.Empty(), RegexpNode.Empty()) => RegexpNode.Empty()
        | (RegexpNode.Range(a), RegexpNode.Range(b)) => RegexpNode.Range(a | b)
        | (a, b) => RegexpNode.Union(a, b)
      }
      def build_closure(expr)
      {
        | RegexpNode.Empty() => expr
        | e => RegexpNode.Closure(e);
      }
      def is_range(expr)
      {
        | RegexpNode.Range() => true
        | _ => false
      }
      
      def build_regexp(node)
      {
        | Node.Union(a, b)        => build_union(build_regexp(a), build_regexp(b))
        | Node.Concat(a, b)       => build_concat(build_regexp(a), build_regexp(b))
        | Node.Optional(a)        => build_union(build_regexp(a), RegexpNode.Empty())
        | Node.Closure(a)         => build_closure(build_regexp(a))
        | Node.PositiveClosure(a) =>
          def tmp = build_regexp(a);
          build_concat(tmp, build_closure(tmp))
        | Node.Intersection(a, b) =>
          def first = build_regexp(a);
          def second = build_regexp(b);
          unless (is_range(first)) error(a, int_err);
          unless (is_range(second)) error(b, int_err);
          if (is_range(first) && is_range(second))
          {
            def ra = (first :> RegexpNode.Range).value;
            def rb = (second :> RegexpNode.Range).value;
            RegexpNode.Range(ra & rb)
          }
          else
            RegexpNode.Empty()
        | Node.Difference(a, b) =>
          def first = build_regexp(a);
          def second = build_regexp(b);
          unless (is_range(first)) error(a, diff_err);
          unless (is_range(second)) error(b, diff_err);
          if (is_range(first) && is_range(second))
          {
            def ra = (first :> RegexpNode.Range).value;
            def rb = (second :> RegexpNode.Range).value;
            RegexpNode.Range(ra & ~rb)
          }
          else
            RegexpNode.Empty()
        | Node.Complement(a) =>
          def expr = build_regexp(a);
          if (is_range(expr))
            RegexpNode.Range(~(expr :> RegexpNode.Range).value)
          else
          {
            error(a, compl_err);
            RegexpNode.Empty()
          }
        | Node.Id(id) =>
          if (PredefinedRegexps.Keys.Contains(id))
            PredefinedRegexps[id]
          else
          {
            error(node, $"Cannot process references to non-subst rules yet: $id");
            RegexpNode.Empty()
          }
        | Node.Range(first, second) =>
          if (first > second)
          {
            error(node, "The first argument of '..' operation must not be greater than the second argument");
            RegexpNode.Empty()
          }
          else
          {
            RegexpNode.Range(CharRange(first, second))
          }
        | Node.String(value) =>
            def f(a, b)
            {
              build_concat(b, RegexpNode.Range(CharRange(a)))
            }
            value.FoldLeft(RegexpNode.Empty(), f)
        | _ => error(node, "Invalid AST"); RegexpNode.Empty()
      }
      def result = ListDictionary();
      foreach (id in rules.Keys)
        result.Add(id, build_regexp(rules[id]));
      result
    }
  }
}