using Nemerle;
using Nemerle.Collections;
using Nemerle.Utility;
using System;
using System.Globalization;
using System.Text;
using SCG = System.Collections.Generic;

namespace Rsdn.ParserTools
{
  public class CharRange
  {
    private class DisjointTree
    {
      public mutable First : DisjointTree;
      public mutable Second : DisjointTree;
    }
    
    private static unicodeRanges : IConstListDictionary[UnicodeCategory, CharRange];
    private static emptyRange : CharRange;
    private static fullRange : CharRange;
    private static eofRange : CharRange;
    
    private hasCharacter : RleArray[bool];
    
    static this()
    {
      unicodeRanges = CreateUnicodeRanges();
      emptyRange = CharRange();
      fullRange = ~emptyRange;
      eofRange = CharRange(-1, -1);
    }
    
    public this()
    {
      hasCharacter = RleArray(-1, char.MaxValue :> int);
    }
    
    public this(c : char)
    {
      this();
      hasCharacter.Fill(true, c :> int, 1);
    }
    
    public this(range : ListRange)
    {
      this();
      hasCharacter.Fill(true, range);
    }
    
    public this(first : int, last : int)
    {
      this(ListRange(first, last + 1));
    }
    
    public this(prototype : RleArray[bool])
    {
      hasCharacter = prototype;
    }

    public static Empty : CharRange
    {
      get
      {
        emptyRange
      }
    }
    
    public static Full : CharRange
    {
      get
      {
        fullRange
      }
    }
    
    public static Eof : CharRange
    {
      get
      {
        eofRange
      }
    }
    
    public IsEmpty : bool
    {
      get
      {
        hasCharacter.IsPlain() && !hasCharacter[-1]
      }
    }
    
    public IsFull : bool
    {
      get
      {
        hasCharacter.IsPlain() && hasCharacter[-1]
      }
    }
    
    public IncludedCharsCount : int
    {
      get
      {
        CountRanges(IncludedRanges)
      }
    }
    
    public IncludedRanges : SCG.IEnumerable[ListRange]
    {
      get
      {
        hasCharacter.RangesOf(true)
      }
    }
    
    public ExcludedCharsCount : int
    {
      get
      {
        CountRanges(ExcludedRanges)
      }
    }
    
    public ExcludedRanges : SCG.IEnumerable[ListRange]
    {
      get
      {
        hasCharacter.RangesOf(false)
      }
    }
    
    public Contains(c : char) : bool
    {
      hasCharacter[c :> int]
    }
    
    public Contains(charRange : CharRange) : bool
    {
      if (this != CharRange.Empty)
      {
        //charRange.IncludedRanges.TrueForAll(
        //  subrange => !hasCharacter[subrange.First] || !hasCharacter.IsPlain(subrange));
        def intersection = CharRange.Intersection(this, charRange);
        intersection.Equals(charRange) 
      }
      else
      {
        false
      } 
    }
    
    [OverrideObjectEquals]
    public Equals(charRange : CharRange) : bool
    {
      this == charRange
    }
    
    public override GetHashCode() : int
    {
      hasCharacter.GetHashCode()
    }
    
    public override ToString() : string
    {
      def sb = StringBuilder();
      _ = sb.Append("[");
      foreach (range in IncludedRanges)
      {
        if (range.First == -1)
        {
          _ = sb.Append(string.Empty);
        }
        else
        {
          if (range.Count == 1)
            _ = sb.Append(range.First :> char);
          else
          {
            _ = sb.Append(range.First :> char);
            _ = sb.Append("-");
            _ = sb.Append((range.Last - 1) :> char);
          }
        }
      }
      _ = sb.Append("]");
      sb.ToString();
    }
    
    public static GetUnicodeRange(category : UnicodeCategory) : CharRange
    {
      unicodeRanges[category]
    }
    
    public static Complement(charRange : CharRange) : CharRange
    {
      def complement(subrange : ListRange, result : CharRange)
      {
        result.hasCharacter.Fill(true, subrange);
        result
      }
      charRange.ExcludedRanges.Fold(CharRange(), complement)
    }
    
    public static Intersection(a : CharRange, b : CharRange) : CharRange
    {
      def result = CharRange();
      foreach (subrange in a.IncludedRanges)
      {
        result.hasCharacter.Fill(true, subrange);
      }
      foreach (subrange in b.ExcludedRanges)
      {
        result.hasCharacter.Fill(false, subrange);
      }
      result
    }
    
    public IntersectWith(charRange : CharRange) : void
    {
      foreach (subrange in charRange.ExcludedRanges)
      {
        this.hasCharacter.Fill(false, subrange);
      }
    }
    
    public static Union(a : CharRange, b : CharRange) : CharRange
    {
      def result = CharRange();
      foreach (subrange in a.IncludedRanges)
      {
        result.hasCharacter.Fill(true, subrange);
      }
      foreach (subrange in b.IncludedRanges)
      {
        result.hasCharacter.Fill(true, subrange);
      }
      result
    }
    
    public UnitWith(charRange : CharRange) : void
    {
      foreach (subrange in charRange.IncludedRanges)
      {
        this.hasCharacter.Fill(true, subrange);
      }
    }
    
    public UnitWith(range : ListRange) : void
    {
      this.hasCharacter.Fill(true, range);
    }
    
    public ComplementTo(range : CharRange) : CharRange
    {
      def result = CharRange();
      foreach (subrange in range.IncludedRanges)
      {
        result.hasCharacter.Fill(true, subrange);
      }
      foreach (subrange in this.IncludedRanges)
      {
        result.hasCharacter.Fill(false, subrange);
      }
      result
    }
    
    public static @~ (charRange : CharRange) : CharRange
    {
      Complement(charRange)
    }
    
    public static @| (a : CharRange, b : CharRange) : CharRange
    {
      Union(a, b)
    }
    
    public static @- (a : CharRange, b : CharRange) : CharRange
    {
      b.ComplementTo(a);
    }
    
    public static @& (a : CharRange, b : CharRange) : CharRange
    {
      Intersection(a, b)
    }
    
    public static @== (a : CharRange, b : CharRange) : bool
    {
      a.hasCharacter.Equals(b.hasCharacter);
    }
    
    public static @!= (a : CharRange, b : CharRange) : bool
    {
      !(a == b)
    }
    
    public static SplitToDisjoint(ranges : SCG.IEnumerable[CharRange]) : SCG.IEnumerable[CharRange]
    {
      def partition = RleArray(-1, Char.MaxValue :> int);
      mutable nullValue = DisjointTree();
      partition.Fill(nullValue);
      
      foreach (range in ranges)
      {
        def includeRange(x : DisjointTree) : DisjointTree
        {
          if (x.First != null)
            x.First
          else
          {
            def t = DisjointTree();
            x.First = t;
            t
          }
        }
        def excludeRange(x : DisjointTree) : DisjointTree
        {
          if (x.Second != null)
            x.Second
          else
          {
            def t = DisjointTree();
            x.Second = t;
            when (x.Equals(nullValue)) nullValue = t;
            t
          } 
        }
        foreach (subRange in range.IncludedRanges)
          partition.Convert(subRange, includeRange);
        foreach (subRange in range.ExcludedRanges)
          partition.Convert(subRange, excludeRange);
      }
      def result = Vector();
      foreach (value when !value.Equals(nullValue) in partition.GetDifferentValues())
        result.Add(CharRange(partition.SetOf(value)));
      result
    }
    
    /*
      Returns summary length of enumerated ranges 
    */
    private CountRanges(ranges : SCG.IEnumerable[ListRange]) : int
    {
      ranges.Fold(0, (range, count) => { count + range.Count })
    }
    
    private static CreateUnicodeRanges() : IConstListDictionary[UnicodeCategory, CharRange]
    {
      /*
        
      */
      def categorize(current : int, last : int, lastCat, ranges)
      {
        def c = current :> char;
        def cat = char.GetUnicodeCategory(c);
        def range = ranges[lastCat];
        if (c != char.MaxValue)
        {
          if (c != char.MinValue && cat != lastCat)
          {
            range.hasCharacter.Fill(true, last, current - last);
            categorize(current + 1, current, cat, ranges);
          }
          else
          {
            categorize(current + 1, last, cat, ranges);
          }
        }
        else
        {
          range.hasCharacter.Fill(true, last, (char.MaxValue :> int) - last);
        }
      }
      
      def ranges = ListDictionary();
      foreach (category in Enum.GetNames(typeof(UnicodeCategory)))
      {
        def category = Enum.Parse(typeof(UnicodeCategory), category) :> UnicodeCategory;
        def range = CharRange();
        ranges.Add(category, range);
      }
      categorize(0, 0, UnicodeCategory.ClosePunctuation, ranges);
      ranges
    }
  }
}
