using Nemerle.Collections;
using Nemerle.Utility;
using System;
using System.Text;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools.Finite
{
  public module DfaBuilder
  {
    public FromNfa(nfa : Nfa) : Dfa[DfaState]
    {
      def edge(state : NfaState, cond : CharRange) : IConstListSet[NfaState]
      {
        nfa.Function.GetTransitions(state, cond)
      }
      def edges(states : SCG.IEnumerable[NfaState], cond : CharRange) : ConstListSet[NfaState]
      {
        states.Fold(
          ConstListSet([]),
          (state, acc) => ConstListSet(edge(state, cond).ConcatenateLazy(acc)));
      }
      def null_closures(states : ConstListSet[NfaState]) : ConstListSet[NfaState]
      {
        def result = HashSet();
        def queue = Queue();
        
        foreach (state in states)
          queue.Push(state);
          
        while (!queue.IsEmpty)
        {
          def state = queue.Take();
          unless (result.Contains(state))
          {
            result.Add(state);
            def nullTransitions = edge(state, CharRange.Empty);
            foreach (s when !result.Contains(s) in nullTransitions)
              queue.Push(s);
          }
        }
        ConstListSet(result)
      }
      def null_closure(state : NfaState)
      {
        null_closures(ConstListSet([state]))
      }
      def find_state(states, state)
      {
        if (states.Contains(state))
          (true, states.GetOriginal(state))
        else
          (false, state)
      }
      
      def states = Queue.[ConstListSet[NfaState]]();
      def dfa = Dfa.[ConstListSet[NfaState]]();
      def dfaStates = ListSet();
      def first = null_closure(nfa.Initial);
      states.Enqueue(first);
      
      while (!states.IsEmpty)
      {
        def dfaState = states.Take();        
        dfaStates.Add(dfaState);
        def conditions = dfaState.Fold(
          ConstListSet([]),
          (state, acc) => ConstListSet(nfa.Function.GetConditions(state).ConcatenateLazy(acc)));
        def disjointConds = CharRange.SplitToDisjoint(conditions.Filter(x => x != CharRange.Empty));
        
        foreach (cond in disjointConds)
        {
          def s = null_closures(edges(dfaState, cond));
          def (isCreated, state) = find_state(dfaStates, s);
          when (!isCreated)
          {
            states.Push(state);
            dfaStates.Add(state);
          }
          dfa.AddTransition(dfaState, cond, state);
        }
      }
      foreach (state in dfaStates)
        dfa.Add(state);
      dfa.Initial = first;
      
      def map = SCG.Dictionary();
      def result = Dfa();
      foreach (state in dfa.States)
      {
        def dfaState = DfaState(state.Map(nfaState => nfaState.RegexpID).Filter(x => x != string.Empty));
        when (state.Contains(x => x.IsFinal))
        {
          dfaState.IsFinal = true;
        }
        map.Add(state, dfaState);
        result.Add(dfaState);
      }
      result.Initial = map[dfa.Initial];
      
      foreach ((init, cond, final) in dfa.Transitions)
      {
        result.AddTransition(map[init], cond, map[final]);
      }
      
      result
    }
   
    public Minimize(dfa : Dfa[DfaState]) : Dfa[DfaState]
    {
      // build initial partition
      //
      def states = ListSet(dfa.States);
      def initialClassMap' = ListMultiDictionary();
      foreach (state in dfa.States)
      {
        def stateClass =
          if (state.IsFinal)
            option.Some(ConstListSet(state.RegexpIDs))
          else
            option.None();
        initialClassMap'.Add(stateClass, state);
      }
      def initialClassMap = ListMultiDictionary();
      foreach (key in initialClassMap'.Keys)
        initialClassMap.AddMany(object(), initialClassMap'[key]);
      
      // build transition map 
      //
      def transitionMap = ListDictionary();
      def nullState = DfaState([]);
      foreach (state in dfa.States)
      {
        def charMap = RleArray(-1, char.MaxValue :> int);
        charMap.Fill(nullState);
        transitionMap.Add(state, charMap);
      }
      foreach ((from, range, to) in dfa.Transitions)
        foreach (subrange in range.IncludedRanges)
          transitionMap[from].Fill(to, subrange);
      
      // build partition when possible
      //
      def nullClass = object();
      def build_partitions(classMap)
      {
        def stateMap = ListDictionary();
        foreach (key in classMap.Keys)
          foreach (val in classMap[key])
            stateMap.Add(val, key);
        def classMap' = ListMultiDictionary();
        foreach (group in classMap.Keys)
        {
          def localMap = ListMultiDictionary();
          foreach (state in classMap[group])
          {
            def charMap = RleArray(-1, char.MaxValue :> int);
            charMap.Fill(nullClass);
            def charMap' = transitionMap[state];
            foreach (subrange in charMap'.AllRanges)
            {
              def val = charMap'[subrange.First];
              when (!val.Equals(nullState))
                charMap.Fill(stateMap[val], subrange);
            }
            localMap.Add(charMap, state);
          }
          foreach (key in localMap.Keys)
          {
            def group' = object();
            foreach (state in localMap[key])
              classMap'.Add(group', state);
          }
        }
        if (classMap.Keys.Count == classMap'.Keys.Count)
          classMap
        else
          build_partitions(classMap')
      }
      
      // build minimized DFA
      //
      def result = Dfa();
      def partition = build_partitions(initialClassMap);
      def stateMap = ListDictionary();
      foreach (key in partition.Keys)
        foreach (value in partition[key])
          stateMap.Add(value, key);
      foreach (group in partition.Keys)
      {
        def state = partition[group][0];
        result.Add(state);
      }
      foreach ((from, range, to) when stateMap.Keys.Contains(from) in dfa.Transitions)
      {
        def from' = partition[stateMap[from]][0];
        def to' = if (stateMap.Keys.Contains(to)) partition[stateMap[to]][0] else to;
        result.AddTransition(from', range, to');
      }
      
      result
    } 
  }
}