using Nemerle.Utility;
using System;
using System.Collections.Generic;
using System.Text;

namespace Rsdn.ParserTools.Finite
{
  type TransitionDiagram = NfaState * TransitionFunction[NfaState] * NfaState;
  
  public module NfaBuilder
  {
    public FromRegexps(regexps : IConstListDictionary [string, RegexpNode]) : Nfa
    {
      Combine(regexps.Map((id, regexp) => FromRegexp(regexp, id)))
    }
    
    /*
      Combines set of Nfas to one
      
      ToDo: what if there is only one Nfa?
    */
    private Combine(nfas : IEnumerable[Nfa]) : Nfa
    {
      def initial = NfaState();
      def finals = ListSet();
      def function = TransitionFunction();
      
      foreach (nfa in nfas)
      {
        function.Add((initial, CharRange.Empty, nfa.Initial));
        function.Append(nfa.Function);
        finals.AddMany(nfa.Finals);
      }
      
      Nfa(initial, function, finals)
    }
    
    /*
      Builds Nfa for regular expression
    */
    public FromRegexp(regexp : RegexpNode, id : string) : Nfa
    {
      // first we build transition diagram
      def (init, function, final) = FromRegexp(regexp);
      // then transform it to Nfa
      def nfa = Nfa(init, function, ListSet([final]));
      
      // assign regexp id to each final state of Nfa
      foreach (final in nfa.Finals)
      {
        final.RegexpID = id;
      }
      nfa
    }
    
    /*
      Builds transition diagram for regular expression
    */
    private FromRegexp(regexp : RegexpNode) : TransitionDiagram
    {
      | Empty           => FromRange(CharRange.Empty)
      | Range (r)       => FromRange(r)
      | Union (e1, e2)  => Unit(FromRegexp(e1), FromRegexp(e2))
      | Concat (e1, e2) => Concatenate(FromRegexp(e1), FromRegexp(e2))
      | Closure (e)     => Enclose(FromRegexp(e))
    }
    
    private FromRange(range : CharRange) : TransitionDiagram
    {
      def initial = NfaState();
      def final = NfaState();
      def function = TransitionFunction();
      
      function.Add((initial, range, final));
      
      (initial, function, final)
    }
    
    private Unit(
      first : TransitionDiagram,
      second : TransitionDiagram) : TransitionDiagram
    {
      def initial = NfaState();
      def final = NfaState();
      def function = TransitionFunction();
      def (fInit, fFunc, fFinal) = first;
      def (sInit, sFunc, sFinal) = second;
      
      function.Append(fFunc);
      function.Append(sFunc);
      function.Append([
        (initial, CharRange.Empty, fInit),
        (initial, CharRange.Empty, sInit),
        (fFinal, CharRange.Empty, final),
        (sFinal, CharRange.Empty, final)
      ]);
      
      (initial, function, final)
    }
    
    private Concatenate(
      first : TransitionDiagram,
      second : TransitionDiagram) : TransitionDiagram
    {
      def function = TransitionFunction();
      def (fInit, fFunc, fFinal) = first;
      def (sInit, sFunc, sFinal) = second;
      
      function.Append(fFunc);
      function.Append(sFunc);
      function.Add((fFinal, CharRange.Empty, sInit));
      
      (fInit, function, sFinal)
    }
    
    private Enclose(diagram : TransitionDiagram) : TransitionDiagram
    {
      def initial = NfaState();
      def final = NfaState();
      def function = TransitionFunction();
      def (i, func, f) = diagram;
      
      function.Append(func);
      function.Append([
        (initial, CharRange.Empty, final),
        (initial, CharRange.Empty, i),
        (f, CharRange.Empty, final),
        (f, CharRange.Empty, i)
      ]);
     
     (initial, function, final)       
    }
  }
}