using Nemerle;
using Nemerle.Collections;
using Nemerle.Imperative;
using Nemerle.Utility;
using System;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools.Grammar
{  
  public class Production
  {
    public Producer : Nonterminal;
    public Sentence : list[Symbol * string * TextLocator];
    public Tag      : string;
    public Locator  : TextLocator;
    hash = 0;

    public this(
      producer : Nonterminal,
       sentence : list[Symbol * string * TextLocator],
       tag : string,
       locator : TextLocator)
    {
      this.Producer = producer;
      this.Sentence = sentence;
      this.Tag = tag;
      this.Locator = locator;
      hash = Producer.GetHashCode();
      foreach ((symbol, _, _) in Sentence)
      {
        hash <<= 2;
        hash ^= symbol.GetHashCode()
      }
    }

    [OverrideObjectEquals]
    public Equals(other : Production) : bool
    {
      this == other
    }
    
    public override GetHashCode() : int
    {
      hash
    }

    public static SameTags(first : Production, second : Production) : bool
    {
      def compare(a : Symbol * string * TextLocator, b : Symbol * string * TextLocator)
      {
        if (a[1] != b[1]) 0 else -1
      }
      List.Compare(first.Sentence, second.Sentence, compare) == 0 &&
      first.Tag == second.Tag;
    }
    
    public static @==(a : Production, b : Production) : bool
    {
      def compare(a, b)
      {
        if (a[0].Equals(b[0])) 0 else -1
      }
      object.ReferenceEquals(a, b) ||
      (a.Producer.Equals(b.Producer) &&
      List.Compare(a.Sentence, b.Sentence, compare) == 0)
    }
    
    public static @!=(a : Production, b : Production) : bool
    {
      !(a == b)
    }
    
    public override ToString() : string
    {
      mutable s = $"$Producer";
      when (Tag != null) s += $"[$Tag]";
      s += " -> ";
      def v = Vector();
      foreach ((symbol, tag, _) in Sentence)
      {
        def t = $"$symbol" + if (tag != null) $"[$tag]" else string.Empty;
        v.Add(t);
      }
      s + string.Join(" ", v.ToArray())
    }
  }
  
  public class Grammar
  {
    public Terminals : IConstListSet[Terminal];
    public Nonterminals : IConstListSet[Nonterminal];
    public Productions : IConstListMultiDictionary[Symbol, Production];
    public ProductionSet : IConstListSet[Production];
    public First : IConstListMultiDictionary[Symbol, Terminal];
    public Follow : IConstListMultiDictionary[Symbol, Terminal];
    
    public this(
      terminals    : IConstListSet[Terminal],
      nonterminals : IConstListSet[Nonterminal],
      productions  : IConstListMultiDictionary[Symbol, Production])
    {
      this.Terminals = terminals;
      this.Nonterminals = nonterminals;
      this.Productions = productions;
      def productionSet = ListSet();
      this.ProductionSet = productionSet;
      foreach (symbol in productions.Keys)
        productionSet.AddMany(productions[symbol]);
      def first = ListDictionary();
      def get_first_many(symbols, f)
      {
        match (symbols)
        {
          | [] => ListSet([EmptyTerminal.Instance])
          | x :: y =>
            def first = f(x[0]);
            if (first.Contains(EmptyTerminal.Instance))
            {
              def newFirst = ListSet(first.Filter(x => !x.Equals(EmptyTerminal.Instance)));
              newFirst.AddMany(get_first_many(y, f));
              newFirst
            }
            else
              first
        }
      }
      def get_first(symbol)
      {
        if (first.Keys.Contains(symbol))
          first[symbol];
        else
        {
          def set = ListSet();
          first.Add(symbol, set);
          match (symbol)
          {
            | s is Terminal => set.Add(s);
            | _ =>
              foreach (prod in productions[symbol])
                set.AddMany(get_first_many(prod.Sentence, get_first))
          }
          set
        }
      }
      def follow = ListMultiDictionary();
      def visited = HashSet();
      def get_follow(symbol)
      {
        if (visited.Contains(symbol))
          follow[symbol]
        else
        {
          visited.Add(symbol);
          foreach (prod in productions)
          {
            def loop(sentence)
            {
              def fill_current(sentence)
              {
                | x :: y =>
                  when (x[0].Equals(symbol))
                  {
                    def set = get_first_many(y, get_first);
                    follow.AddMany(symbol, set.Filter(x => !x.Equals(EmptyTerminal.Instance)));
                    when (set.Contains(EmptyTerminal.Instance))
                      follow.AddMany(symbol, get_follow(prod.Producer));
                  }
                | _ => ();
              }
              match (sentence)
              {
                | _ :: y => fill_current(sentence); loop(y);
                | [] => ();
              }
            }
            loop(prod.Sentence);
          }
          follow[symbol]
        }
      }
      def firstMap = ListMultiDictionary();
      
      // I found that covariance is not supported :(
      //
      foreach (symbol in terminals)
        firstMap.AddMany(symbol, get_first(symbol));
      foreach (symbol in nonterminals)
        firstMap.AddMany(symbol, get_first(symbol));
        
      this.First = firstMap;
      foreach (symbol in terminals)
        _ = get_follow(symbol);
      foreach (symbol in nonterminals)
        _ = get_follow(symbol);
      this.Follow = follow;
    }
  }
}