using Nemerle;
using Nemerle.Collections;
using Nemerle.Imperative;
using System;
using System.IO;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools.Grammar
{
  public module GrammarUtils
  {    
    public ReportGrammar(grammar : Grammar, writer : TextWriter) : void
    {
      writer.WriteLine($"Terminals ($(grammar.Terminals.Count)):");
      foreach (symbol in grammar.Terminals)
      {
        writer.WriteLine($"\t$symbol");
      }
      writer.WriteLine($"\nNonterminals ($(grammar.Nonterminals.Count)):");
      foreach (symbol in grammar.Nonterminals)
      {
        writer.WriteLine($"\t$symbol");
      }
      writer.WriteLine($"\nProductions ($(grammar.Productions.Count)):");
      foreach (nonterm in grammar.Productions.Keys)
      {
        foreach (prod in grammar.Productions[nonterm])
        {
          writer.WriteLine($"\t$prod")
        }
      }
      def set_to_str(set)
      {
        def v = Vector();
        foreach (item in set)
          v.Add(item.ToString());
        string.Join(", ", v.ToArray());
      }
      writer.WriteLine("\nFIRST:");
      foreach (symbol in grammar.Nonterminals)
      {
        def str = set_to_str(grammar.First[symbol]);
        writer.WriteLine($"\t$symbol: {$str}");
      }
      writer.WriteLine("\nFOLLOW:");
      foreach (symbol in grammar.Nonterminals)
      {
        def str = set_to_str(grammar.Follow[symbol]);
        writer.WriteLine($"\t$symbol: {$str}");
      }
      foreach (symbol in grammar.Terminals)
      {
        def str = set_to_str(grammar.Follow[symbol]);
        writer.WriteLine($"\t$symbol: {$str}");
      }
    }
    
    public ItemAsString(item : LrItem) : string
    {
      def v = Vector();
      def loop(sentence, index)
      {
        when (index == item.Index) v.Add(".");
        match (sentence)
        {
          | []     => ();
          | x :: y => v.Add(x[0].ToString()); loop(y, index + 1)
        }
      }
      loop(item.Production.Sentence, 0);
      def r = string.Join(" ", v.ToArray());
      $"$(item.Production.Producer) -> $r"
    }
    
    public ReportSlrSystem(slr : SlrSystem, writer : TextWriter) : void
    {
      writer.WriteLine($"States ($(slr.Transitions.Keys.Count)):");
      foreach (state in slr.States)
      {
        def index = slr.States.IndexOf(state);
        writer.WriteLine($"\tI$index:");
        foreach (item in state.Items)
        {
          def str = ItemAsString(item);
          writer.WriteLine($"\t\t$str");
        }
      }
      writer.WriteLine("\nTransitions:");
      foreach (src in slr.States)
      {
        def srcIndex = slr.States.IndexOf(src);
        foreach ((symbol, dest) in slr.Transitions[src])
        {
          def destIndex = slr.States.IndexOf(dest);
          writer.WriteLine($"\t$symbol: I$srcIndex -> I$destIndex");
        }
      }
    }
    
    public ReportLalrSystem(lalr : LalrSystem, writer : TextWriter) : void
    {
      def item_to_string(item, la)
      {
        def str = ItemAsString(item);
        def v = Vector();
        foreach (term in la)
          v.Add(term.ToString());
        def lastr = string.Join(" | ", v.ToArray());
        $"$str, $lastr";
      }
      writer.WriteLine($"States ($(lalr.Transitions.Keys.Count)):");
      foreach (state in lalr.States)
      {
        def index = lalr.States.IndexOf(state);
        writer.WriteLine($"\tI$index:");
        foreach (item in state.Items)
        {
          def str = item_to_string(item, lalr.Lookahead[index][item]);
          writer.WriteLine($"\t\t$str");
        }
      }
      writer.WriteLine("\nTransitions:");
      foreach (src in lalr.States)
      {
        def srcIndex = lalr.States.IndexOf(src);
        foreach ((symbol, dest) in lalr.Transitions[src])
        {
          def destIndex = lalr.States.IndexOf(dest);
          writer.WriteLine($"\t$symbol: I$srcIndex -> I$destIndex");
        }
      }
    }
    
    public BuildClosure(grammar : Grammar, state : LrItemSet) : LrItemSet
    {
      def set = ListSet();
      def queue = Queue();
      foreach (item in state.Items)
        queue.Enqueue(item);
      while (queue.Count > 0)
      {
        def item = queue.Dequeue();
        when (set.Contains(item)) continue;
        set.Add(item);
        match (item.CurrentSymbol)
        {
          | nonterm is Nonterminal =>
            foreach (prod in grammar.Productions[nonterm])
              queue.Enqueue(LrItem(prod));
          | _ => ()
        }
      }
      LrItemSet(set)
    }
    
    public BuildSlrSystem(grammar : Grammar) : SlrSystem
    {
      def states = ListSet();
      def transitions = ListMultiDictionary();
      def queue = Queue();
      def closure = FunctionCache(BuildClosure);
      def firstItem = LrItem(grammar.Productions[StartNonterminal.Instance][0]);
      def firstStateRaw = LrItemSet(ConstListSet([firstItem]));
      queue.Enqueue(closure.Get(grammar, firstStateRaw));
      while (!queue.IsEmpty)
      {
        def curr = queue.Dequeue();
        when (states.Contains(curr)) continue;
        states.Add(curr);
        def nextStates = ListMultiDictionary();
        foreach (item when item.HasNext in curr.Items)
          nextStates.Add(item.CurrentSymbol, item.GetNext());
        foreach (symbol in nextStates.Keys)
        {
          def nextStateRaw = LrItemSet(nextStates[symbol]);
          def nextState = closure.Get(grammar, nextStateRaw);
          transitions.Add(curr, (symbol, nextState));
          queue.Enqueue(nextState);
        }
      }
      SlrSystem(states, transitions)
    }
    
    public BuildLalrSystem(grammar : Grammar) : LalrSystem
    {
      // builds closure for specified raw state and transitions for
      // undeterministic automaton
      //
      def build_closure(grammar, state)
      {
        def set = ListSet();
        def transitions = ListSet();
        def queue = Queue();
        foreach (item in state.Items)
          queue.Enqueue(item);
        while (queue.Count > 0)
        {
          def item = queue.Dequeue();
          when (set.Contains(item)) continue;
          set.Add(item);
          match (item.CurrentSymbol)
          {
            | nonterm is Nonterminal =>
              foreach (prod in grammar.Productions[nonterm])
              {
                def newItem = LrItem(prod);
                transitions.Add((item, newItem));
                queue.Enqueue(LrItem(prod));
              }
            | _ => ()
          }
        }
        (LrItemSet(set), transitions)
      }
      
      def get_first(item)
      {
        def loop(list, accum)
        {
          match (list)
          {
            | [] => accum.Add(EmptyTerminal.Instance); accum
            | (x, _, _) :: y =>
              def first = grammar.First[x];
              if (!first.Contains(EmptyTerminal.Instance))
              {
                accum.AddMany(first);
                accum
              }
              else
              {
                accum.AddMany(first.Filter(fun(s) { !s.Equals(EmptyTerminal.Instance) }  ));
                loop(y, accum)
              }
          }
        }
        loop(item.GetNext().Current, ListSet())
      }
      
      def states = ListSet();
      def transitions = ListMultiDictionary();
      def subTransitions = ListMultiDictionary();
      
      def queue = Queue();
      def closure = FunctionCache(build_closure);
      def firstItem = LrItem(grammar.Productions[StartNonterminal.Instance][0]);
      def firstStateRaw = LrItemSet(ConstListSet([firstItem]));
      def (firstState, firstTransitions) = closure.Get(grammar, firstStateRaw);
      def spontaneous = ListMultiDictionary();
      def first = FunctionCache(get_first);      
      def add_closure_trans(iter, state)
      {
        foreach ((src, dest) in iter)
        {
          def firstset = first.Get(src);
          if (firstset.Contains(EmptyTerminal.Instance))
          {
            subTransitions.Add((state, src), (state, dest));
            def newSymbols = firstset.Filter(x => !x.Equals(EmptyTerminal.Instance));
            spontaneous.AddMany((state, dest), newSymbols);
          }
          else
            spontaneous.AddMany((state, dest), firstset);
        }
      }
      add_closure_trans(firstTransitions, firstState);
      queue.Enqueue(firstState);
      
      // builds both deterministic automaton and initial lookahead information
      //
      while (!queue.IsEmpty)
      {
        def curr = queue.Dequeue();
        when (states.Contains(curr)) continue;
        states.Add(curr);
        def nextStates = ListMultiDictionary();
        def tmpTransitions = ListMultiDictionary();
        foreach (item when item.HasNext in curr.Items)
        {
          def nextItem = item.GetNext();
          nextStates.Add(item.CurrentSymbol, nextItem);
          tmpTransitions.Add(item.CurrentSymbol, (item, nextItem));
        }
        foreach (symbol in nextStates.Keys)
        {
          def nextStateRaw = LrItemSet(nextStates[symbol]);
          def (nextState, closureTrans) = closure.Get(grammar, nextStateRaw);
          transitions.Add(curr, (symbol, nextState));
          queue.Enqueue(nextState);
          foreach ((src, dest) in tmpTransitions[symbol])
            subTransitions.Add((curr, src), (nextState, dest));
          when (!states.Contains(nextState))
            add_closure_trans(closureTrans, nextState);
        }
      }

      def fill_lookahead()
      {
        def lookahead = array(states.Count);
        for (mutable i = 0; i < lookahead.Length; ++i)
          lookahead[i] = ListMultiDictionary();
        def queue = Queue();
        queue.Enqueue(firstState, firstItem, ConstListSet([EofTerminal.Instance]));
        foreach ((state, item) in spontaneous.Keys)
          queue.Enqueue(state, item, spontaneous[state, item]);
        while (queue.Count > 0)
        {
          def (state, item, symbols) = queue.Dequeue();
          def stateIndex = states.IndexOf(state);
          def lasymbols = lookahead[stateIndex][item];
          def newSymbols = ConstListSet(symbols.Filter(x => !lasymbols.Contains(x)));
          when (newSymbols.Count == 0) continue;
          lookahead[stateIndex].AddMany(item, newSymbols);
          foreach ((deststate, destitem) in subTransitions[state, item])
            queue.Enqueue(deststate, destitem, newSymbols);
        }
        lookahead
      }
      def lookahead = fill_lookahead();
      LalrSystem(states, transitions, lookahead)
    }
  }
}