using Nemerle;
using Nemerle.Collections;
using Nemerle.Imperative;
using Nemerle.Utility;
using System;
using System.IO;
using System.IO.Compression;
using System.Xml;
using System.Xml.Xsl;
using Rsdn.ParserTools.Bnf;
using Rsdn.ParserTools.Grammar;
using Rsdn.ParserTools.Finite;


namespace Rsdn.ParserTools
{
  public module Generator
  {
    public CreateLRTable(grammar : Grammar, lalr : LalrSystem) : LRTable
    {
      def symbolsCount = grammar.Terminals.Count + grammar.Nonterminals.Count;
      def result = LRTable(lalr.States.Count, symbolsCount);
      def index_of_symbol(symbol : Symbol)
      {
        if (symbol is Terminal)
          grammar.Terminals.IndexOf(symbol :> Terminal)
        else
          grammar.Nonterminals.IndexOf(symbol :> Nonterminal) + grammar.Terminals.Count
      }
      def add_reduction(state, symbol, prod)
      {
        when (result.Actions[state, symbol] == 0)
        {
          if (prod == 0)
            result.Actions[state, symbol] = 3
          else
          {
            result.Actions[state, symbol] = 2;
            result.Productions[state, symbol] = prod;
          }
        }
      }
      def add_shift(from, to, symbol)
      {
        if (symbol < grammar.Terminals.Count)
        {
          result.Actions[from, symbol] = 1;
          result.States[from, symbol] = to;
        }
		else
          result.States[from, symbol] = to;
      }
      
      // adds reduction cells
      //
      foreach (state in lalr.States)
      {
        def stateIndex = lalr.States.IndexOf(state);
        foreach (item when item.CurrentSymbol == null in state.Items)
        {
          def prodIndex = grammar.ProductionSet.IndexOf(item.Production);
          foreach (symbol in lalr.Lookahead[stateIndex][item])
            add_reduction(stateIndex, index_of_symbol(symbol), prodIndex);
        }
      }
      
      // adds shift cells
      //
      foreach (from in lalr.Transitions.Keys)
      {
        foreach ((symbol, to) in lalr.Transitions[from])
          add_shift(
            lalr.States.IndexOf(from),
            lalr.States.IndexOf(to),
            index_of_symbol(symbol));
      }
      result
    }
    
    public CreateDfas(unit : Node.Root, log : IErrorLog) : IConstListDictionary[string, DfaInfo]
    {
      def dfas = ListDictionary();
      foreach (id in unit.lexers.Keys)
        dfas.Add(id, DfaInfo.CreateDfaInfo(unit.lexers[id], log));
      dfas
    }
    
    /*
        Returns id of regexp accepted by given state, or null if state is not final.
     */
    public GetStateName(lexer : Node.Section, state : DfaState) : string
    {
      if (!state.IsFinal)
        null
      else
      {
        def ids = state.RegexpIDs.ToList();
        match (List.Sort(ids, (a, b) => lexer.body.Keys.IndexOf(b).CompareTo(lexer.body.Keys.IndexOf(a))))
        {
          | h :: _ => h
          | _ => null
        }
      }
    }
    
    public CreateParserNode(
      doc : XmlDocument,
      grammar : Grammar,
      unit : Node.Root,
      log : IErrorLog,) : XmlElement
    {
      def root = doc.CreateElement("lr");
      
      def error(loc, message)
      {
        log.Message(Error(loc, ErrorLevel.Error, message))
      }
      
      def get_copier_index(production)
      {
        def f(list, curr, index)
        {
          match (list : list[Symbol * string * TextLocator])
          {
            | [] => curr
            | (_, "_", _) :: tail =>
              if (curr == -1) f(tail, index, index + 1) else -1
            | (_, tag, _) :: _ when tag != null => -1
            | (_ is Nonterminal, _, _) :: t =>
              if (curr == -1) f(t, index, index + 1) else -1
            | _ :: t => f(t, curr, index + 1)
          }
        }
        match (production.Sentence)
        {
          | (_, tag, _) :: [] when tag == null => 0
          | s => f(s, -1, 0)
        }
      }
      
      def walk_production(production)
      {
        def option = if (production.Tag != null) production.Tag else production.Producer.Name;
        def set_rule_and_index(element)
        {
          element.SetAttribute("rule", production.ToString());
          element.SetAttribute("index", grammar.ProductionSet.IndexOf(production).ToString());
        }
        def fill_args(element)
        {
          def constructor = unit.ast[option];
          element.SetAttribute("option", option);
          when (constructor.isfun) element.SetAttribute("isfunction", "isfunction");
          mutable i = 0;
          def tagMap = ListDictionary();
          foreach ((_, tag, loc) in production.Sentence)
          {
            when (tag != null)
            {
              tagMap.Add(tag, i);
              when (!constructor.args.Keys.Contains(tag))
                log.Message(Error(
                  loc, ErrorLevel.Error,
                  $"Unknown constructor parameter reference: $tag"));
            }
            ++i;
          }
          foreach (tag in constructor.args.Keys)
          {
            def field = constructor.args[tag];
            def child = doc.CreateElement("arg");
            _ = element.AppendChild(child);
            def index = if (tagMap.Keys.Contains[tag]) tagMap[tag] else -1;
            child.SetAttribute("type", field.fieldType);
            child.SetAttribute("index", index.ToString());
          }
        }
        
        def copier_index = get_copier_index(production);
        if (production.Sentence == [])
        {
          def element = doc.CreateElement("e-reduction");
          set_rule_and_index(element);
          element
        }
        else if (copier_index != -1)
        {
          def element = doc.CreateElement("copy");
          set_rule_and_index(element);
          element.SetAttribute("copyindex", copier_index.ToString());
          element
        }
        else if (unit.ast.Keys.Contains(option) && copier_index == -1)
        {
          def element = doc.CreateElement("reduction");
          set_rule_and_index(element);
          fill_args(element);
          element
        }
        else
        {
          error(TextLocator(null, TextRange.Empty), $"Error in grammar: $production");
          null
        }
      }
      
      def grammarElement = doc.CreateElement("grammar");
      _ = root.AppendChild(grammarElement);
      foreach (prod in grammar.ProductionSet)
      {
        def prodElement = doc.CreateElement("production");
        prodElement.SetAttribute("length", prod.Sentence.Length.ToString());
        def symbolIndex =
          grammar.Nonterminals.IndexOf(prod.Producer) +
          grammar.Terminals.Count;
        prodElement.SetAttribute("symbol", symbolIndex.ToString());
        prodElement.SetAttribute("index", grammar.ProductionSet.IndexOf(prod).ToString());
        _ = grammarElement.AppendChild(prodElement);
      }
      
      def tokenTypes = doc.CreateElement("tokentypes");
      _ = root.AppendChild(tokenTypes);
      foreach (term in grammar.Terminals)
      {
        def element = doc.CreateElement("tokentype");
        _ = tokenTypes.AppendChild(element);
        element.SetAttribute("name", term.Name);
        element.SetAttribute("value", grammar.Terminals.IndexOf(term).ToString());
      }
      
      def reductionsElement = doc.CreateElement("reductions");
      _ = root.AppendChild(reductionsElement);
      foreach (prod in grammar.ProductionSet)
      {
        def element = walk_production(prod);
        when (element != null)
          _ = reductionsElement.AppendChild(element);
      }
      
      foreach (id in unit.ast.Keys)
      {
        def constructor = unit.ast[id];
        when (constructor.isfun)
        {
          def consElement = doc.CreateElement("constructor");
          consElement.SetAttribute("id", id);
          foreach (fieldId in constructor.args.Keys)
          {
            def field = constructor.args[fieldId];
            def fieldElement = doc.CreateElement("field");
            fieldElement.SetAttribute("id", fieldId);
            fieldElement.SetAttribute("type", field.fieldType);
            _ = consElement.AppendChild(fieldElement);
          }
          _ = root.AppendChild(consElement);
        }
      }
      
      root
    }
    
    public EscapeNemerleString(text : string) : string
    {
      def escape(text, acc)
      {
        if (text == string.Empty)
          acc
        else
        {
          def h = text[0];
          def t = text.Substring(1);
          def v = match (h)
          {
            | '\\' => "\\\\"
            | '"'  => "\\\""
            | '\'' => "\\\'"
            | _ => h.ToString()
          }
          escape(t, acc + v)
        }
      }
      escape(text, string.Empty)
    }
    
    public CreateLexerNode(
      doc : XmlDocument,
      dfas : IConstListDictionary[string, DfaInfo],
      unit : Node.Root,
      _ : IErrorLog) : XmlElement
    {
      def root = doc.CreateElement("fa");
      root.SetAttribute("cachetext", "cachetext");
      
      def nameGroups = doc.CreateElement("names-group");
      _ = root.AppendChild(nameGroups);
      foreach (id in dfas.Keys)
      {
        def dfaInfo = dfas[id];
        def lexer = unit.lexers[id];
        def namesElement = doc.CreateElement("names");
        foreach (state in dfaInfo.Dfa.States)
        {
          def nameElement = doc.CreateElement("name");
          def acceptedName = GetStateName(lexer, state);
          nameElement.InnerText = if (acceptedName == null) "Error" else acceptedName;
          _ = namesElement.AppendChild(nameElement);
        }
        def keywordMap = ListMultiDictionary();
        foreach (keyword in dfaInfo.Keywords.Keys)
        {
          def terminal = dfaInfo.Keywords[keyword];
          foreach (text in terminal.TextList)
            keywordMap.Add(terminal.TargetId, (text, keyword));
        }
        
        namesElement.SetAttribute("num", dfas.Keys.IndexOf(id).ToString());
        foreach (terminal in keywordMap.Keys)
        {
          def terminalElement = doc.CreateElement("terminal");
          terminalElement.SetAttribute("id", terminal);
          foreach ((text, keyword) in keywordMap[terminal])
          {
            def keywordElement = doc.CreateElement("keyword");
            keywordElement.SetAttribute("id", keyword);
            keywordElement.SetAttribute("text", EscapeNemerleString(text));
            _ = terminalElement.AppendChild(keywordElement);
          }
          _ = namesElement.AppendChild(terminalElement);
        }
        _ = nameGroups.AppendChild(namesElement);
      }
      
      def added = ListSet();
      def enumElement = doc.CreateElement("enum");
      _ = root.AppendChild(enumElement);
      foreach (lexerId in unit.lexers.Keys)
      {
        def lexer = unit.lexers[lexerId];
        foreach (id when !added.Contains(id) in lexer.body.Keys)
        {
          added.Add(id);
          def itemElement = doc.CreateElement("item");
          itemElement.InnerText = id;
          _ = enumElement.AppendChild(itemElement);
        }
      }
      
      root
    }

    public CreateXml(
      grammar : Grammar,
      unit : Node.Root,
      dfas : IConstListDictionary[string, DfaInfo],
      log : IErrorLog) : XmlDocument
    {
      unless (unit.header.Keys.Contains("namespace"))
      {
        log.Message(Error(unit.loc, ErrorLevel.Error, "Required parameter 'namespace' was not defined"));
        log.SetPanic();
        return null;
      }
      
      def doc = XmlDocument();
      def root = doc.CreateElement("root");
      _ = doc.AppendChild(root);
      root.SetAttribute("namespace", unit.header["namespace"].value);
      root.SetAttribute(
        "token",
        if (unit.header.Keys.Contains("tokenClass")) unit.header["tokenClass"].value else "Token");
      root.SetAttribute(
        "tokentype",
        if (unit.header.Keys.Contains("tokenTypeEnum")) unit.header["tokenTypeEnum"].value else "TokenType");
     
      when (unit.parser != null)
      {
        def parserClassName = if (unit.header.Keys.Contains("parserClass")) unit.header["parserClass"].value else "ParserHelper";
        root.SetAttribute("parser", parserClassName);
        root.SetAttribute(
          "node",
          if (unit.header.Keys.Contains("node")) unit.header["node"].value else "Node");
        root.SetAttribute(
          "node-null",
          if (unit.header.Keys.Contains("null")) unit.header["null"].value else "Null");
        root.SetAttribute(
          "builder",
          if (unit.header.Keys.Contains("builder")) unit.header["builder"].value else "NodeBuilder");
        def lr = CreateParserNode(doc, grammar, unit, log);
        lr.SetAttribute("resource-name", "Parser.bin");
        _ = root.AppendChild(lr);
      }
      
      def get_bool_param(id, defaultValue)
      {
        if (!unit.header.Keys.Contains(id))
          defaultValue
        else 
          match (unit.header[id].value)
          {
            | "true" => true
            | "false" => false
            | some =>
              log.Message(Error(
                unit.header[id].loc,
                ErrorLevel.Warning,
                $"Parameter '$id' has invalid value '$some'. Assuming default value '$defaultValue'"));
              defaultValue
          }
      }
      
      def generateLexer = if (unit.lexers.Keys.Count == 0)
        false
      else get_bool_param("generateLexer", true);
        
      def generateLexerFacade = get_bool_param("generateLexerFacade", true);

      when (generateLexer)
      {
        def lexerClassName = if (unit.header.Keys.Contains("tokenizerhelper")) unit.header["tokenizerhelper"].value else "LexerHelper";
        root.SetAttribute("tokenizerhelper", lexerClassName);
        root.SetAttribute(
          "tokenmanager",
          if (unit.header.Keys.Contains("tokenManager")) unit.header["tokenManager"].value else "TokenManager");
        def fa = CreateLexerNode(doc, dfas, unit, log);
        fa.SetAttribute("resourcefile", "Lexer.bin");
        fa.SetAttribute("facade",
          if (unit.header.Keys.Contains("lexerFacadeClass")) unit.header["lexerFacadeClass"].value else "Lexer");
        fa.SetAttribute("generateFacade", generateLexerFacade.ToString());
        _ = root.AppendChild(fa);
      }
      
      def usings = [];
      
      foreach (usingEntry in usings)
      {
        def usingElement = doc.CreateElement("using");
        usingElement.InnerText = usingEntry;
        _ = root.AppendChild(usingElement);
      }
      
      doc
    }

    public WriteLRTable(grammar : Grammar, table : LRTable, stream : Stream) : void
    {
      def compressor = DeflateStream(stream, CompressionMode.Compress);
      def writer = StreamWriter(compressor);
      def height = table.States.GetUpperBound(0) + 1;
      def width = table.States.GetUpperBound(1) + 1;
      writer.WriteLine(height);
      for (mutable i = 0; i < height; ++i)
      {
        def actionRow = Vector();
        def gotoRow = Vector();
        for (mutable j = 0; j < grammar.Terminals.Count; ++j)
        {
          match (table.Actions[i, j])
          {
            | 1 => actionRow.Add((j, table.Actions[i, j], table.States[i, j]));
            | 2 => actionRow.Add((j, table.Actions[i, j], table.Productions[i, j]));
            | 3 => actionRow.Add((j, table.Actions[i, j], 0));
            | _ => ()
          }
        }
        for (mutable j = grammar.Terminals.Count; j < width; ++j)
          when (table.States[i, j] != -1)
            gotoRow.Add((j, table.States[i, j]));
        writer.WriteLine(actionRow.Count);
        foreach ((term, action, p) in actionRow)
        {
          writer.WriteLine(term);
          writer.WriteLine(action);
          writer.WriteLine(p);
        }
        writer.WriteLine(gotoRow.Count);
        foreach ((sym, p) in gotoRow)
        {
          writer.WriteLine(sym);
          writer.WriteLine(p);
        }
      }
      writer.Close();
    }
    
    public WriteDfa(
      dfas : IConstListDictionary[string, DfaInfo],
      stream : Stream) : void
    {
      def get_transitions_map(transitions) : IConstListMultiDictionary[DfaState, CharRange * DfaState]
      {
        def map = ListMultiDictionary();
        foreach ((from, range, to) in transitions)
          map.Add(from, (range, to));
        map
      }
        
      def compressor = DeflateStream(stream, CompressionMode.Compress);
      def writer = StreamWriter(compressor);
      
      writer.WriteLine(dfas.Keys.Count);
      foreach (id in dfas.Keys)
      {
        def dfa = dfas[id].Dfa;
        def states = ListSet(dfa.States);
        def transitions = get_transitions_map(dfa.Transitions);
        writer.WriteLine(states.Count);
        foreach (from in states)
        {
          def ranges = Vector();
          foreach ((range, to) in transitions[from])
          {
            def toIndex = states.IndexOf(to);
            foreach (listRange in range.IncludedRanges)
              ranges.Add((listRange.First, listRange.Last, toIndex));
          }
          ranges.Sort((a, b) => a[0].CompareTo(b[0]));
          writer.WriteLine(ranges.Count);
          foreach ((rangeFrom, rangeTo, destState) in ranges)
          {
            writer.WriteLine(rangeFrom);
            writer.WriteLine(rangeTo - 1);
            writer.WriteLine(destState);
          }
        }
      }
      writer.Close();
    }
  }
}