using Nemerle;
using Nemerle.Collections;
using Nemerle.IO;
using Nemerle.Imperative;
using System;
using System.IO;
using System.Xml;
using System.Xml.Xsl;
using Rsdn.ParserTools;
using Rsdn.ParserTools.Bnf;
using Rsdn.ParserTools.Grammar;


def transform_xml(doc, writer, xslt)
{
  def transform = XslCompiledTransform();
  def transformReader = XmlTextReader.Create(
    typeof(Generator).Assembly.GetManifestResourceStream(typeof(Generator), xslt));
  transform.Load(transformReader);
  transformReader.Close();
  def xmlWriter = XmlTextWriter.Create(writer, transform.OutputSettings);
  transform.Transform(doc, xmlWriter);
  xmlWriter.Close();
}

def generate(log, url, reportUrl)
{
  def reader = StreamReader(url);
  def lexer = Lexer(reader, url);
  def parser = Parser();
  def tree = parser.Parse(lexer, log);
  when (log.Panic || tree == null) return false;
  
  def unit = BnfUtils.ApplySubst(tree :> Node.Root, log);
  
  def grammar = if (unit.parser != null) BnfUtils.BuildGrammar(unit, log) else null;
  when (log.Panic) return false;
  def lalr = if (grammar != null) GrammarUtils.BuildLalrSystem(grammar) else null;
  when (log.Panic) return false;

  when (reportUrl != null && grammar != null)
  {
    def reportWriter = StreamWriter($"$url.report");
    GrammarUtils.ReportGrammar(grammar, reportWriter);
    GrammarUtils.ReportLalrSystem(lalr, reportWriter);
    reportWriter.WriteLine();
    reportWriter.Close();
  }

  def dfas = Generator.CreateDfas(unit, log);
  def table = if (grammar != null) Generator.CreateLRTable(grammar, lalr) else null;
  when (log.Panic) return false;
  
  def doc = Generator.CreateXml(grammar, unit, dfas, log);
  when (log.Panic) return false;
  
  if (log.HasErrors)
    false
  else
  {
    //doc.Save($"$url.xml");
    def parserWriter = StreamWriter("parsing.n");
    transform_xml(doc, parserWriter, "xslt.Parser.n.xslt");
    parserWriter.Close();

    when (doc.GetElementsByTagName("lr").Count > 0)
    {
      def parserBinWriter = FileStream("Parser.bin", FileMode.Create);
      Generator.WriteLRTable(grammar, table, parserBinWriter);
      parserBinWriter.Close();
    }
 
    when (doc.GetElementsByTagName("fa").Count > 0)
    {
      def lexerBinWriter = FileStream("Lexer.bin", FileMode.Create);
      Generator.WriteDfa(dfas, lexerBinWriter);
      lexerBinWriter.Close();
    }
    true
  }
}

def cmdLine = CommandLineUtils.GetCommandLine();
unless (cmdLine.Keys.Contains("source") && cmdLine["source"].Count == 1)
{
  Console.WriteLine("usage: \n\tnpg source=source_file_name [report[=report_file_name]]");
  Environment.Exit(0);
}
def url = cmdLine["source"][0];
def reportUrl = if (cmdLine["report"].Count == 0)
  null
else
{
  when (cmdLine["report"].Count > 1)
    Console.Error.WriteLine("Warning: more than one entry of report parameter found. Using the first one.");
  if (cmdLine["report"][0] == string.Empty)
    "$url.report"
  else
    cmdLine["report"][0]
};

def start = DateTime.Now.Ticks;
def log = SimpleErrorLog();
def success = 
try
{
  generate(log, url, reportUrl);
}
catch
{
  | e =>
    log.Message(Error(TextLocator(url, TextRange.Empty), ErrorLevel.Error, $"Exception thrown:\n$e"));
    false
}


if (success)
{
  def end = DateTime.Now.Ticks;
  log.Write(Console.Error);
  def elapsed = (end - start) / 10000;
  print("elapsed $(elapsed)ms\n");
}
else
{
  log.Write(Console.Error);
}
Environment.Exit(if (success) 0 else 1);