﻿using Nemerle;
using Nemerle.Collections;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;
using Nemerle.Imperative;
using Nemerle.Macros;
using System;
using System.Text;
using Rsdn.ParserTools;
using Rsdn.ParserTools.Finite;
using BNF = Rsdn.ParserTools.Bnf;
using SCG = System.Collections.Generic;


namespace Rsdn.ParserTools
{
  module BnfMacroHelper
  {
    public DefineTokenClass(env : GlobalEnv) : void
    {
      def tokenBuilder = env.Define(<[decl:
          [Record]
          public class Token
          {
            public TokenType : TokenType;
            public Locator : TextLocator;
            public Text : string
          }
        ]>);
        tokenBuilder.Compile();
    }
    
    public DefineTokenTypeClass(
      env : GlobalEnv,
      unit : BNF.Node.Root) : void
    {
      def variants = Vector();
      def added = ListSet();
      foreach (id in unit.lexers.Keys)
      {
        def lexer = unit.lexers[id];
        foreach (ruleId when !added.Contains(ruleId) in lexer.body.Keys)
        {
          added.Add(ruleId);
          variants.Add(<[decl: | $(Macros.UseSiteSymbol(ruleId) : name) ]>);
        }
      }
      variants.AddRange([ <[decl: | Eof ]>, <[decl: | Error ]> ]);
      def variantList = variants.ToList();
      def tokenTypeBuilder = env.Define(<[decl:
        public enum TokenType : int
        {
          ..$(variantList)
        }
      ]>);
      tokenTypeBuilder.Compile();
    }
    
    DefineGetTokenMethod(builder : TypeBuilder) : void
    {
      def tokenName = Macros.UseSiteSymbol("Token");
      builder.Define(<[decl:
        public GetToken() : $(tokenName : name)
        {
          if (ch == -1)
            $(tokenName : name)(TokenType.Eof, TextLocator(Url, TextRange(row, column, row, column)), string.Empty)          
          else
          {
            def firstRow = row;
            def firstColumn = column;
            mutable lastRow = row;
            mutable lastColumn = column;
            _ = strBuilder.Remove(0, strBuilder.Length);
            mutable state = 0;
            while (true)
            {
              lastRow = row;
              lastColumn = column;
              def nextState = GetNextState(state);
              when (nextState == -1) break;
              state = nextState;
              when (ch == -1) break;
              _ = strBuilder.Append(ch :> char);
              get_next_char();
            }
            def tokenType = stateNames[Mode][state];
            when (tokenType == TokenType.Error)
              get_next_char();
            def text = strBuilder.ToString();
            $(tokenName : name)(
              TransformToken(tokenType, text),
              TextLocator(Url, TextRange(firstRow, firstColumn, lastRow, lastColumn)),
              text);
          }
        }
      ]>);
    }
    
    public GetStateName(lexer : BNF.Node.Section, state : DfaState) : string
    {
      if (!state.IsFinal)
        null
      else
      {
        def ids = state.RegexpIDs.ToList();
        match (List.Sort(ids, (a, b) => lexer.body.Keys.IndexOf(b).CompareTo(lexer.body.Keys.IndexOf(a))))
        {
          | h :: _ => h
          | _ => null
        }
      }
    }
    
    DefineStateNamesField(
      builder : TypeBuilder,
      dfas : IConstListDictionary[string, DfaInfo]) : void
    {
      mutable exprList = [];
      foreach (lexerId in dfas.Keys)
      {
        def dfa = dfas[lexerId];
        mutable stateList = [];
        foreach (state in dfa.Dfa.States)
        {
          def name = GetStateName(dfa.Lexer, state);
          if (name != null)
            stateList = <[ TokenType.$(Macros.UseSiteSymbol(name) : name) ]> :: stateList
          else
            stateList = <[ TokenType.Error ]> :: stateList;
        }
        stateList = stateList.Reverse();
        exprList = <[ array[ ..$stateList ] ]> :: exprList;
      }
      exprList = exprList.Reverse();
      def body = <[ array[ ..$exprList ] ]>;
      //PrettyPrint.PrintExpr(option.None(), body);
      builder.Define(<[decl:
        stateNames : array[array[TokenType]] = $body;
      ]>);
    }
    
    DefineTransformTokenMethod(
      builder : TypeBuilder,
      dfas : IConstListDictionary[string, DfaInfo]) : void
    {
      mutable exprList = [];
      foreach (lexerId in dfas.Keys)
      {
        def dfa = dfas[lexerId];
        mutable tokenList = [];
        def map = ListMultiDictionary();
        foreach (keywordId in dfa.Keywords.Keys)
        {
          def keyword = dfa.Keywords[keywordId];
          foreach (text in keyword.TextList)
            map.Add(keyword.TargetId, (keywordId, text));
        }
        foreach (terminal in map.Keys)
        {
          mutable textList = [];
          foreach ((keyword, text) in map[terminal])
            textList = <[case:
              | $(text : string) => TokenType.$(Macros.UseSiteSymbol(keyword) : name)
            ]> :: textList;
          textList = <[case: | _ => token ]> :: textList;
          textList = textList.Reverse();
          tokenList = <[case:
            | TokenType.$(Macros.UseSiteSymbol(terminal) : name) =>
              match(text) { ..$textList } ]> :: tokenList;
        }
        tokenList = <[case: | _ => token ]> :: tokenList;
        tokenList = tokenList.Reverse();
        def index = dfas.Keys.IndexOf(lexerId);
        exprList = <[case: | $(index : int) => match(token) { ..$tokenList } ]> :: exprList;
      }
      exprList = <[case: | _ => token ]> :: exprList;
      exprList = exprList.Reverse();
      def body = <[ match (Mode) { ..$exprList } ]>;
      //PrettyPrint.PrintExpr(option.None(), body);
      builder.Define(<[decl:
        TransformToken(token : TokenType, text : string) : TokenType { $body } ]>);
    }
    
    DefineGetNextStateMethod(
      builder : TypeBuilder,
      dfas : IConstListDictionary[string, DfaInfo]) : void
    {
      def get_transitions_map(transitions) : IConstListMultiDictionary[DfaState, CharRange * DfaState]
      {
        def map = ListMultiDictionary();
        foreach ((from, range, to) in transitions)
          map.Add(from, (range, to));
        map
      }
      
      def get_search_tree(ranges,l, u)
      {
        if (l > u)
          <[ -1 ]>
        else
        {
          def i = (l + u) / 2;
          def (from, to, dest) = ranges[i];
          def lessBranch = get_search_tree(ranges, l, i - 1);
          def greaterBranch = get_search_tree(ranges, i + 1, u);
          <[
            if (ch < $(from : int))
              $lessBranch
            else if (ch >= $(to : int))
              $greaterBranch
            else
              $(dest : int)
          ]>
        }
      }
      
      mutable exprList = [];
      foreach (lexerId in dfas.Keys)
      {
        def dfa = dfas[lexerId];
        def states = ListSet(dfa.Dfa.States);
        def transitions = get_transitions_map(dfa.Dfa.Transitions);
        mutable stateList = [];
        foreach (from in states)
        {
          def fromIndex = states.IndexOf(from);
          def ranges = RleArray(-1, char.MaxValue :> int);
          ranges.Fill(-1);
          foreach ((range, to) in transitions[from])
          {
            def toIndex = states.IndexOf(to);
            foreach (listRange in range.IncludedRanges)
              ranges.Fill(toIndex, listRange);
          }
          def smoothRanges = ranges.Clone() :> RleArray[int];
          match (smoothRanges.RangesCount)
          {
            | x when x < 16 => ()
            | x when x < 32 => smoothRanges.Smooth(2);
            | x when x < 48 => smoothRanges.Smooth(4);
            | _ => smoothRanges.Smooth(6);
          }
          when (smoothRanges.RangesCount > 25)
            smoothRanges.Smooth(6);
          def ranges' = Vector();
          mutable exceptList = [];
          foreach (range in smoothRanges.AllRanges)
          {
            ranges'.Add(range.First, range.Last, smoothRanges[range.First]);
            def val = smoothRanges[range.First];
            foreach (sharpRange in ranges.RangesAt(range))
            {
              def sharpVal = ranges[sharpRange.First];
              when (sharpVal != val)
              {
                for (mutable i = sharpRange.First; i < sharpRange.Last; ++i)
                  exceptList = <[ ( $(i : int), $(sharpVal : int) ) ]> :: exceptList;
              }
            }
          }
          def smoothRangesCount = Vector(smoothRanges.AllRanges).Count;
          def rangesCount = Vector(ranges.AllRanges).Count;
          def stateExpr = get_search_tree(ranges', 0, ranges'.Count - 1);
          def stateExpr' =
            if (exceptList != [])
            {
              def exceptSymbol = Macros.NewSymbol();
              builder.Define(<[decl:
                static $(exceptSymbol : name) : ListDictionary[int, int] =
                  ListDictionary.[int, int]( array[ ..$exceptList ] );
              ]>);
              <[
                if ($(exceptSymbol : name).Keys.Contains(ch))
                  $(exceptSymbol : name)[ch]
                else
                  $stateExpr
              ]>
            }
            else
              stateExpr;
          stateList = <[case: | $(fromIndex : int) => $stateExpr' ]> :: stateList;
        }
        stateList = <[case: | _ => -1 ]> :: stateList;
        stateList = stateList.Reverse();
        def dfaExpr = <[ match (state) { ..$stateList } ]>;
        def mode = dfas.Keys.IndexOf(lexerId);
        exprList = <[case: | $(mode : int) => $dfaExpr ]> :: exprList; 
      }
      exprList = <[case: | _ => -1 ]> :: exprList;
      exprList = exprList.Reverse();
      def methodBody = <[ match (Mode) { ..$exprList } ]>;
      //PrettyPrint.PrintExpr(option.None(), methodBody);
      builder.Define(<[decl: GetNextState(state : int) : int { $methodBody } ]>);
    }
    
    public DefineLexerClass(
      env : GlobalEnv,
      unit : BNF.Node.Root,
      dfas : IConstListDictionary[string, DfaInfo],
      log : IErrorLog) : void
    {
      def builder = env.Define(<[decl:
        public class LexelHelper
        {
          mutable row : int;
          mutable column : int;
          reader : System.IO.TextReader;
          mutable cr : bool;
          mutable ch : int;
          strBuilder : StringBuilder;
          public Url : string;
          public mutable Mode : int;
          
          public this(reader : System.IO.TextReader, url : string)
          {
            this.reader = reader;
            this.Url = url;
            strBuilder = StringBuilder();
            ch = 0;
            column = -1;
            get_next_char();
          }
        }
      ]>);
      def tokenName = Macros.UseSiteSymbol("Token");
      if (env.Manager.IsIntelliSenseMode)
      {
        builder.Define(<[decl:
          public GetToken() : $(tokenName : name)
          {
            $(tokenName : name)(TokenType.Eof, TextLocator.Empty, string.Empty)
          }
        ]>)
      }
      else
      {        
        builder.Define(<[decl:        
          public get_next_char() : void
          {
            if (ch == 13)
            {
              cr = true;
              ++row;
              column = 0;
            }
            else if (ch == 10)
            {
              column = 0;
              when (!cr) ++row;
              cr = false;
            }
            else
            {
              ++column;
              cr = false;
            }
            ch = reader.Read();
          }
        ]>);
        DefineGetTokenMethod(builder);
        DefineGetNextStateMethod(builder, dfas);
        DefineStateNamesField(builder, dfas);
        DefineTransformTokenMethod(builder, dfas);
      }
      builder.Compile();
    }
  
    GetTokenSequence(token : Token) : SCG.IEnumerable[Token]
    {
      match (token)
      {
        | null => yield null;
        | Token.RoundGroup(_) =>
          yield Token.BeginRound();
          mutable first = true;
          foreach (token' :> Token.LooseGroup in token)
          {
            unless (first)
              yield Token.Comma();
            first = false;
            foreach (token'' in GetTokenSequence(token'))
              yield token'';
          }
          yield Token.EndRound();
        | Token.BracesGroup(_) =>
          yield Token.BeginBrace(false);
          foreach (token' :> Token.LooseGroup in token)
          {
            foreach (token'' in GetTokenSequence(token'))
              yield token'';
            yield Token.Semicolon(false);
          }
          yield Token.EndBrace(false);
        | Token.SquareGroup(child) =>
          yield Token.BeginSquare();
          mutable first = true;
          foreach (token' :> Token.LooseGroup in token)
          {
            unless (first)
              yield Token.Comma();
            first = false;
            foreach (token'' in GetTokenSequence(token'))
              yield token'';
          }
          yield Token.EndSquare();
        | Token.QuoteGroup(child) =>
          yield Token.BeginQuote();
          foreach (token' :> Token.LooseGroup in token)
          {
            foreach (token'' in GetTokenSequence(token'))
              yield token'';
            yield Token.Semicolon(false);
          }
          yield Token.EndQuote();
        | Token.LooseGroup() as t =>
          foreach (token' :> Token in t)
            foreach (token'' :> Token in GetTokenSequence(token'))
              yield token''
        | other =>
          yield other;
      }
    }
    
    public GetTokens(token : Token) : SCG.IEnumerable[Token]
    {
      foreach (token' in GetTokenSequence(token))
          yield token';
    }
    
    public ConvertLocation(loc : Location) : TextLocator
    {
      TextLocator(loc.File, TextRange(loc.Line, loc.Column, loc.EndLine, loc.EndColumn))
    }
  }
}