
using Nemerle.Collections;
using Nemerle.Imperative;
using System;
using System.IO;
using System.IO.Compression;
using System.Text;
using Rsdn.ParserTools;

namespace Rsdn.ParserTools.Test
{

  public interface ITokenSource
  {
    GetNextToken() : Token;
    EvalToken(token : Token, log : IErrorLog) : Node;
  }
  
  public interface INodeBuilder
  {
    Binary(loc : TextLocator, first : Node, op : Node, second : Node) : Node;
  Prefix(loc : TextLocator, op : Node, expr : Node) : Node;
  Function(loc : TextLocator, id : Node, args : Node) : Node;
  List(loc : TextLocator, head : Node, tail : Node) : Node;
  
  }

  public class ParserHelper
  {
    private enum Action : byte
    {
      | Error
      | Shift
      | Reduce
      | Accept
    }

    static actionTable : array[array[int * Action * int]];
    static gotoTable : array[array[int * int]];
    mutable panic = false;

    static ProductionInfo(prod : int) : int * int
    {
      | 0 => (11, 2)
      | 1 => (12, 2)
      | 2 => (12, 1)
      | 3 => (13, 3)
      | 4 => (13, 3)
      | 5 => (13, 1)
      | 6 => (14, 3)
      | 7 => (14, 3)
      | 8 => (14, 1)
      | 9 => (15, 1)
      | 10 => (15, 1)
      | 11 => (15, 1)
      | 12 => (15, 1)
      | 13 => (16, 3)
      | 14 => (17, 4)
      | 15 => (18, 3)
      | 16 => (18, 1)
      | 17 => (19, 1)
      | _ => throw Exception("Internal parser error")
    }

    static TokenToInt(token : TokenType) : int
    {
      | TokenType.Eof => 0
      | TokenType.Identifier => 1
      | TokenType.Number => 2
      | TokenType.Space => 3
      | TokenType.Comma => 4
      | TokenType.Minus => 5
      | TokenType.Plus => 6
      | TokenType.Star => 7
      | TokenType.Slash => 8
      | TokenType.LeftParen => 9
      | TokenType.RightParen => 10
      | _ => throw Exception("Internal parser error")
    }

    static ReduceProduction(loc : TextLocator, builder : INodeBuilder, prod : int, stack : array[Node]) : Node
    {
      match (prod)
      {
        // copy: S_ -> Negative Eof
        | 0 =>
          stack[0]
  
        // reduction: Negative{Prefix} -> Minus{op} Additive{expr}
        | 1 =>
          
        builder.Prefix(loc, stack[0], stack[1])
      
        // copy: Negative{Prefix} -> Additive
        | 2 =>
          stack[0]
  
        // reduction: Additive{Binary} -> Additive{first} Plus{op} Multiplicative{second}
        | 3 =>
          
        builder.Binary(loc, stack[0], stack[1], stack[2])
      
        // reduction: Additive{Binary} -> Additive{first} Minus{op} Multiplicative{second}
        | 4 =>
          
        builder.Binary(loc, stack[0], stack[1], stack[2])
      
        // copy: Additive{Binary} -> Multiplicative
        | 5 =>
          stack[0]
  
        // reduction: Multiplicative{Binary} -> Multiplicative{first} Star{op} Primitive{second}
        | 6 =>
          
        builder.Binary(loc, stack[0], stack[1], stack[2])
      
        // reduction: Multiplicative{Binary} -> Multiplicative{first} Slash{op} Primitive{second}
        | 7 =>
          
        builder.Binary(loc, stack[0], stack[1], stack[2])
      
        // copy: Multiplicative{Binary} -> Primitive
        | 8 =>
          stack[0]
  
        // copy: Primitive -> Parenthesized
        | 9 =>
          stack[0]
  
        // copy: Primitive -> FunctionCall
        | 10 =>
          stack[0]
  
        // copy: Primitive -> Identifier
        | 11 =>
          stack[0]
  
        // copy: Primitive -> Number
        | 12 =>
          stack[0]
  
        // copy: Parenthesized -> LeftParen Negative RightParen
        | 13 =>
          stack[1]
  
        // reduction: FunctionCall{Function} -> Identifier{id} LeftParen Args{args} RightParen
        | 14 =>
          
        builder.Function(loc, stack[0], stack[2])
      
        // reduction: Args{List} -> Arg{head} Comma Args{tail}
        | 15 =>
          
        builder.List(loc, stack[0], stack[2])
      
        // reduction: Args{List} -> Arg{head}
        | 16 =>
          
        builder.List(loc, stack[0], Node.Null(loc))
      
        // copy: Arg -> Negative
        | 17 =>
          stack[0]
  
        | _ => throw Exception("Internal parser error")
      }
    }

    static ToBool(node : Node) : bool
    {
      | Node.Null() => false
      | _ => true
    }

    static this()
    {
      def stream = typeof(ParserHelper).Assembly.GetManifestResourceStream(typeof(ParserHelper), "Parser.bin");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def stateCount = Int32.Parse(reader.ReadLine());
      actionTable = array(stateCount);
      gotoTable = array(stateCount);
      for (mutable i = 0; i < stateCount; ++i)
      {
        def cellCount = Int32.Parse(reader.ReadLine());
        def row = array(cellCount);
        for (mutable j = 0; j < cellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def action = Int32.Parse(reader.ReadLine()) :> Action;
          def param = Int32.Parse(reader.ReadLine());
          row[j] = (symbol, action, param);
        }
        actionTable[i] = row;
        def gotoCellCount = Int32.Parse(reader.ReadLine());
        def gotoRow = array(gotoCellCount);
        for (mutable j = 0; j < gotoCellCount; ++j)
        {
          def symbol = Int32.Parse(reader.ReadLine());
          def param = Int32.Parse(reader.ReadLine());
          gotoRow[j] = (symbol, param);
        }
        gotoTable[i] = gotoRow;
      }
    }

    public Panic : bool
    {
      get
      {
        return panic;
      }
    }

    public ResetPanic() : void
    {
      panic = false;
    }        

    public Parse(source : ITokenSource, builder : INodeBuilder, log : IErrorLog) : Node
    {
      mutable token;
      def stack = Vector.[int * int * Node * Token * Token]();
      def next_token()
      {
        token = source.GetNextToken();
      }
      def state()
      {
        if (stack.Count > 0)
          stack[stack.Count - 1][0]
        else
          0
      }
      def search_array['a, 'b](e : 'a, arr : array['b], comparer)
      {
        def search(l, u)
        {
          def i = (l + u) / 2;
          if (l > u)
            -1
          else
          {
            match (comparer(e, arr[i]))
            {
              | -1  => search(l, i - 1)
              | 1   => search(i + 1, u)
              | _   => i
            }
          }
        }
        search(0, arr.Length - 1)
      }
      def get_action(state, symbol)
      {
        def row = actionTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        if (col >= 0) (row[col][1], row[col][2]) else (Action.Error, -1)
      }
      def get_goto(state, symbol)
      {
        def row = gotoTable[state];
        def col = search_array(symbol, row, fun(a, b) { a.CompareTo(b[0]) });
        row[col][1]
      }
      def shift(state)
      {
        def symbol = TokenToInt(token.TokenType);
        def data = source.EvalToken(token, log);
        stack.Add((state, symbol, data, token, token));
        next_token()
      }
      def reduce(prodIndex)
      {
        def (nonterm, len) = ProductionInfo(prodIndex);
        def get_data(i, acc)
        {
          if (i < 0)
            acc
          else
          {
            def (_, _, data, _, _) = stack[stack.Count - len + i];
            get_data(i - 1, data :: acc)
          }
        }
        def data = get_data(len - 1, []);
        def (firstToken, lastToken) =
          if (len > 0)
            (stack[stack.Count - len][3], stack[stack.Count - 1][4])
          else
            (token, token);
        def loc = TextLocator(firstToken.Locator.Url, TextRange(firstToken.Locator.First, lastToken.Locator.Last));
        def newData = ReduceProduction(loc, builder, prodIndex, data.ToArray());
        stack.RemoveRange(stack.Count - len, len);
        def newState = get_goto(state(), nonterm);
        stack.Add((newState, nonterm, newData, firstToken, lastToken));
      }
      def error()
      {
        def search_state(i)
        {
          if (i < -1)
            i
          else
          {
            def testState = if (i > 0) stack[i][0] else 0;
            def (action, _) = get_action(testState, TokenToInt(token.TokenType));
            match (action)
            {
              | Action.Error => search_state(i - 1)
              | _            => i
            }
          }
        }
        panic = true;
        mutable firstToken = token;
        while (true)
        {
          def stackIndex = search_state(stack.Count - 1);
          if (stackIndex >= -1)
          {
            firstToken = if (stackIndex + 1 < stack.Count)
              stack[stackIndex + 1][3]
            else
              token;
            stack.RemoveRange(stackIndex + 1, stack.Count - stackIndex - 1);
            break;
          }
          else
            next_token();
          when (token.TokenType == TokenType.Eof) break;
        }
        def loc = TextLocator(firstToken.Locator.Url, firstToken.Locator.First, token.Locator.Last);
        log.Message(Error(loc, ErrorLevel.Error, "Invalid syntax"));
        token.TokenType != TokenType.Eof
      }
      def loop()
      {
        def (action, param) = get_action(state(), TokenToInt(token.TokenType));
        match (action)
        {
          | Action.Shift  => shift(param); loop()
          | Action.Reduce => reduce(param); loop()
          | Action.Accept => stack[0][2]
          | Action.Error  => if (error()) loop() else stack[0][2]
        }
      }
      next_token();
      loop()
    }
  }

  public class LexerHelper
  {
    static stateNames : array[array[TokenType]] = array[
      array[
        TokenType.Error,
        TokenType.Identifier,
        TokenType.Number,
        TokenType.Space,
        TokenType.Comma,
        TokenType.Minus,
        TokenType.Plus,
        TokenType.Star,
        TokenType.Slash,
        TokenType.LeftParen,
        TokenType.RightParen,
        TokenType.Identifier,
        TokenType.Number,
        TokenType.Error,
        TokenType.Space,
        TokenType.Number,
        TokenType.Number
      ]
    ];
    static automata : array[array[array[int * int * int]]];
    static initialState : int * int * TokenType = (0, 0, TokenType.Eof);
    reader : TextReader;
    strBuilder : StringBuilder = StringBuilder();
    url : string;
    mutable lastToken : TokenType;
    mutable mode : int;
    mutable ch : int;
    mutable row : int;
    mutable column : int;
    mutable currentState : int;
    mutable cr : bool;
        
    static this()
    {
      def stream = typeof(LexerHelper).Assembly.GetManifestResourceStream(typeof(LexerHelper), "Lexer.bin");
      def decompressor = DeflateStream(stream, CompressionMode.Decompress);
      def reader = StreamReader(decompressor);
      def automataCount = Int32.Parse(reader.ReadLine());
      automata = array(automataCount);
      for (mutable index = 0; index < automataCount; ++index)
      {
        def stateCount = Int32.Parse(reader.ReadLine());
        def states = array(stateCount);
        for (mutable n = 0; n < stateCount; ++n)
        {
          def transitionsCount = Int32.Parse(reader.ReadLine());
          def transitions = array(transitionsCount);
          for (mutable i = 0; i < transitionsCount; ++i)
          {
            def from = Int32.Parse(reader.ReadLine());
            def to = Int32.Parse(reader.ReadLine());
            def state = Int32.Parse(reader.ReadLine());
            transitions[i] = (from, to, state);
          }
          states[n] = transitions;
        }
        automata[index] = states;
      }
      reader.Close();
    }
        
    public this(reader : TextReader, url : string)
    {
       this.reader = reader;
       this.url = url;
       ch = 0;
       column = -1;
       get_next_char();
    }
        
    public static InitialState : object
    {
      get
      {
        return initialState;
      }
    }
        
    get_next_char() : void
    {
      if (ch == 13)
      {
        cr = true;
        ++row;
        column = 0;
      }
      else if (ch == 10)
      {
        column = 0;
        when (!cr) ++row;
        cr = false;
      }
      else
      {
        ++column;
        cr = false;
      }
      ch = reader.Read();
    }
        
    public Mode : int
    {
      get
      {
        return this.mode;
      }
      set
      {
        this.mode = value;
      }
    }
        
    public LastTokenType : TokenType
    {
      get
      {
        return this.lastToken;
      }
    }
        
    public IsEof : bool
    {
      get
      {
        return ch == -1;
      }
    }
        
    public Save() : object
    {
      (currentState, mode, lastToken);
    }
        
    public Reset(state : object, row : int, column : int) : void
    {
      def extState = state :> int * int * TokenType;
      this.currentState = extState[0];
      this.lastToken = extState[2];
      this.mode = extState[1];
      this.row = row;
      this.column = column;
    }
    
    TransformToken(token : TokenType, text : string) : TokenType
    {
      match (mode)
      {
        
        | 0 =>
          match (token)
          {
            
            | _ => token
          }
  
        | _ => token
      }
    }
    
    public GetToken() : Token
    {
      def get_next_state(mode, state, c)
      {
        def s = automata[mode][state];
        def search(l, u)
        {
          if (l > u)
            -1
          else
          {
            def i = (l + u) / 2;
            if (c < s[i][0])
              search(l, i - 1)
            else if (c > s[i][1])
              search(i + 1, u)
            else
              s[i][2]
          }
        }
        search(0, s.Length - 1)
      }
      if (ch == -1)
      {
         Token(
         TokenType.Eof,
          TextLocator(url, TextRange(row, column, row, column)),
          string.Empty)
      }
      else
      {
        def firstRow = row;
        def firstColumn = column;
        mutable lastRow = row;
        mutable lastColumn = column;
        _ = strBuilder.Remove(0, strBuilder.Length);
        mutable lastState = currentState;
        while (true)
        {
          lastRow = row;
          lastColumn = column;
          def nextState = get_next_state(mode, currentState, ch);
          when (nextState == -1) break;
          currentState = nextState;
          when (ch == -1) break;
          lastState = nextState;
          _ = strBuilder.Append(ch :> char);
          get_next_char();
        }
        when (stateNames[mode][currentState] == TokenType.Error)
        {
          get_next_char();
        }
        lastToken = stateNames[mode][currentState];
        if (ch == -1)
        {
          currentState = lastState;
        }
        else
        {
          currentState = 0;
        }
        def text = strBuilder.ToString();
        Token(
          TransformToken(lastToken, text),
          TextLocator(url, TextRange(firstRow, firstColumn, lastRow, lastColumn)),
          text);
      }
    }
  }
  
  public class Lexer : ITokenSource
  {
    helper : LexerHelper;
    
    public this(reader : TextReader, url : string)
    {
      this.helper = LexerHelper(reader, url);
    }
    
    public GetNextToken() : Token
    {
      def loop(token)
      {
        match (token.TokenType)
        {
          | TokenType.Eof => token
          | _ when TokenManager.Filter(token) => token
          | _ => loop(helper.GetToken())
        }
      }
      loop(helper.GetToken())
    }
    
    public EvalToken(token : Token, log : IErrorLog) : Node
    {
      TokenManager.Eval(token, log)
    }
  }
  
  [Record]
  public class Token
  {
    public TokenType : TokenType;
    public Locator : TextLocator;
    public Text : string;

    public this(proto : Token)
    {
      this.TokenType = proto.TokenType;
      this.Text = proto.Text;
      this.Locator = proto.Locator;
    }
        
    public this(proto : Token, tokenType : TokenType)
    {
      this.TokenType = tokenType;
      this.Text = proto.Text;
      this.Locator = proto.Locator;
    }
  }
  
  public enum TokenType
  {
    | Identifier
    | Number
    | Space
    | Comma
    | Minus
    | Plus
    | Star
    | Slash
    | LeftParen
    | RightParen

    | Eof
    | Error
  }

}